import glob

import numpy as np

from mock import *

from atrip.container import guess_container, load
from atrip.media_type import Media, guess_media_type
from atrip import errors

from atrip.media_types.atari_disks import *
from atrip.media_types.apple_disks import *

class TestDiskImage:

    @pytest.mark.parametrize("filename,sector_size,sectors", [
        ("dos_sd_test1.atr", 128, [(1,3), (360,1), (361,8)]),
        ("dos_ed_test1.atr", 128, [(1,3), (360,1), (361,8)]),
    ])
    def test_contiguous_sectors(self, filename, sector_size, sectors):
        pathname = os.path.join(os.path.dirname(__file__), "../samples", filename)
        container = load(pathname)
        media = container.media

        for start, count in sectors:
            # grab sectors
            segment = media.get_contiguous_sectors(start, count)
            # 16 byte ATR header, first byte of sector 1 is byte 16
            assert np.array_equal(segment.container_offset, np.arange(16 + (start - 1) * sector_size, 16 + (start + count - 1)*sector_size))

    def test_ed_sectors(self):
        pathname = os.path.join(os.path.dirname(__file__), "../samples/dos_ed_test1.atr")
        container = load(pathname)
        media = container.media

        # get vtoc + vtoc2
        boot = media.get_sector_list([360,1024])
        assert np.array_equal(boot.container_offset[0:128], np.arange(16 + (360 - 1) * 128, 16 + 360*128))
        assert np.array_equal(boot.container_offset[128:256], np.arange(16 + (1024 - 1) * 128, 16 + 1024*128))


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.WARNING)
    log = logging.getLogger("atrip.media_type")
    log.setLevel(logging.DEBUG)

    def check(pathname):
        print(f"checking {pathname}")
        container = load(pathname)
        print(container.verbose_info)
        media = container.media
        print(media)
        boot = media.get_contiguous_sectors(1, 3)
        print(boot)
        print(boot.container_offset)

    import sys
    import glob
    if len(sys.argv) > 1:
        images = sys.argv[1:]
    else:
        images = sorted(glob.glob(os.path.join(os.path.dirname(__file__), "../samples/", "*")))
    for pathname in images:
        check(pathname)



import os
import sys
from collections import namedtuple

import numpy as np
import wx

from sawx.ui import compactgrid as cg
from ..ui import segment_grid as sg
from ..viewer import SegmentViewer
from .emulator import EmulatorViewerMixin

import logging
log = logging.getLogger(__name__)


from atrip.memory_map import MemoryMap
filename = "./omnivore/templates/atari800.labels"
try:
    labels1 = MemoryMap.from_file(filename)
except:
    log.warning("Can't find local labels for emulator labels test")
else:
    pass #print(labels1.labels)

class LabelTable(cg.VariableWidthHexTable):
    want_col_header = False
    want_row_header = True

    def __init__(self, linked_base):
        self.linked_base = linked_base

        s = linked_base.segment
        super().__init__(s.data, s.style, [], s.origin)
        self.rebuild()

    def size_of_entry(self, d):
        return d[2]

    def parse_table_description(self, desc):
        row_to_label_number = []
        items_per_row = []
        index_of_row = []
        row_of_index = []
        index = 0
        row = 0
        i = 0
        print(desc)
        try:
            self.row_to_label_number = desc.keys()
        except AttributeError:
            self.row_to_label_number = []
        for i in self.row_to_label_number:
            d = desc[i]
            s = self.size_of_entry(d)
            items_per_row.append(s)
            index_of_row.append(index)
            row_of_index.extend([row] * s)
            print(f"row_of_index: {d[0]} index={index}, s={s} {d}") #: {row_of_index}")
            # print(f"index_of_row: {index_of_row}")
            index += s
            row += 1
        self.index_of_row = index_of_row
        self.row_of_index = row_of_index
        self.items_per_row = items_per_row
        self.last_valid_index = index

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        for line in range(start_line, last_line, step):
            yield self.labels[self.row_to_label_number[line]][0]

    def calc_row_label_width(self, view_params):
        return max([view_params.calc_text_width(r) for r in self.get_row_label_text(0, self.num_rows)])

    def calc_last_valid_index(self):
        pass  # calculated in init_table_description

    type_code_fmt = {
        0x00: "02x",  # hex byte
        0x01: "04x",  # hex word
        0x03: "08x",  # hex long
        0x30: "03d",  # dec byte
        0x31: "05d",  # dec word
        0x33: "010d",  # dec long
        0x40: "08b",  # bin byte
        0x41: "016b",  # bin word
        0x43: "032b",  # bin long
    }

    dtype_fmt = [
        None,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint32,
    ]

    def calc_cell_widths(self):
        cell_widths = [1] * self.num_rows
        for row in range(self.num_rows):
            label_num = self.row_to_label_number[row]
            type_code = self.labels[label_num][3]
            cell_widths[row] = len(format(0, self.type_code_fmt[type_code])) // 2  # each cell is 2 chars wide
        # print(cell_widths)
        return cell_widths

    def get_value_style(self, row, col):
        label_num = self.row_to_label_number[row]
        type_code = self.labels[label_num][3]
        bytes_per_col = (type_code & 0x03) + 1
        index = label_num + (col * bytes_per_col)
        dtype = self.dtype_fmt[bytes_per_col]
        try:
            value = int(self.data[index:index + bytes_per_col].view(dtype))
        except IndexError:
            text = ""
            style = 0
        else:
            style = self.style[index]
            text = format(value, self.type_code_fmt[type_code])
        return text, style

    def get_label_at_index(self, index):
        row = self.row_of_index[index]
        return self.labels[self.row_to_label_number[row]][0]

    def prepare_for_drawing(self, start_row, visible_rows, start_cell, visible_cells):
        # for i, item in enumerate(self.labels):
        #     print(i, item)
        pass

    def rebuild(self):
        self.labels = labels1.labels
        self.init_table_description(self.labels)
        self.init_boundaries()
        print(f"new num_rows: {self.num_rows}")


class LabelGridControl(sg.SegmentGridControl):
    default_table_cls = LabelTable
 
    def calc_default_table(self, linked_base):
        table = self.default_table_cls(linked_base)
        self.items_per_row = table.items_per_row
        self.want_row_header = table.want_row_header
        self.want_col_header = table.want_col_header
        return table

    def calc_line_renderer(self):
        print(self.table.items_per_row)
        return cg.VariableWidthLineRenderer(self, 2, self.table.items_per_row, self.table.calc_cell_widths())

    def recalc_view(self):
        log.debug(f"recalc_view: {self}")
        self.table.rebuild()
        self.line_renderer = self.calc_line_renderer()
        super().recalc_view()



class LabelViewer(EmulatorViewerMixin, SegmentViewer):
    name = "labels"

    viewer_category = "Emulator"

    ui_name = "Labels"

    control_cls = LabelGridControl

    @property
    def linked_base_segment_identifier(self):
        return ""



import wx

from ..action import ViewerAction

from sawx.action import SawxAction, SawxListAction, SawxPersistentDictAction
from sawx import errors
from sawx.persistence import get_json_data, save_json_data

import logging
log = logging.getLogger(__name__)


class layout_save(SawxAction):
    name = "Save Layout"

    def save_layout(self, label, layout):
        layout_restore.add(label, layout)

    def perform(self, action_key):
        label = self.editor.frame.prompt("Enter name for layout", "Save Layout")
        if label:
            layout = {}
            self.editor.serialize_session(layout, False)
            self.save_layout(label, layout)


class layout_restore(SawxPersistentDictAction):
    prefix = "layout_restore_"

    json_name = "tilemanager_layouts"

    def get_layout(self, action_key):
        label = self.current_list[self.get_index(action_key)]
        d = self.get_dict()
        return d[label]

    def perform(self, action_key):
        layout = self.get_layout(action_key)
        self.editor.replace_layout(layout)


class layout_save_emu(layout_save):
    name = "Save Emulator Layout"

    def save_layout(self, label, layout):
        layout_restore_emu.add(label, layout)


class layout_restore_emu(layout_restore):
    prefix = "layout_restore_emu_"

    json_name = "tilemanager_emulator_layouts"

    def perform(self, action_key):
        doc = self.editor.document
        doc.pause_emulator()
        try:
            layout_restore.perform(self, action_key)
        finally:
            doc.resume_emulator()


class TileManagerAction(ViewerAction):
    name = "Add Viewer Above"

    @property
    def tilemanger(self):
        return self.editor.control

    @property
    def tile_client(self):
        v = self.viewer
        client = self.tilemanger.find_control(v.control)
        log.debug(f"viewer={v}, client={client}")
        return client


class TileAddAction(TileManagerAction):
    side = None

    def perform(self, action_key):
        client = self.tile_client
        client.split_side(self.side)


class layout_add_viewer_above(TileAddAction):
    name = "Add Viewer Above"
    side = wx.TOP


class layout_add_viewer_below(TileAddAction):
    name = "Add Viewer Below"
    side = wx.BOTTOM


class layout_add_viewer_left(TileAddAction):
    name = "Add Viewer Left"
    side = wx.LEFT


class layout_add_viewer_right(TileAddAction):
    name = "Add Viewer Right"
    side = wx.RIGHT


class TileSidebarAction(TileManagerAction):
    side = None

    def perform(self, action_key):
        client = self.tile_client
        client.minimize(self.side)


class layout_move_to_top_sidebar(TileSidebarAction):
    name = "Move Viewer to Top Sidebar"
    side = wx.TOP


class layout_move_to_bottom_sidebar(TileSidebarAction):
    name = "Move Viewer to Bottom Sidebar"
    side = wx.BOTTOM


class layout_move_to_left_sidebar(TileSidebarAction):
    name = "Move Viewer to Left Sidebar"
    side = wx.LEFT


class layout_move_to_right_sidebar(TileSidebarAction):
    name = "Move Viewer to Right Sidebar"
    side = wx.RIGHT


class layout_remove_viewer_and_close_tile(TileManagerAction):
    name = "Remove Viewer and Tile"

    def perform(self, action_key):
        client = self.tile_client
        client.destroy_thyself()


class layout_remove_viewer_and_keep_tile(TileManagerAction):
    name = "Remove Viewer, Keep Tile"

    def perform(self, action_key):
        client = self.tile_client
        client.clear_child()



# Standard library imports.
import sys
import os

# Major package imports.
import wx
import numpy as np
import json

# Local imports.
from .byte_editor import ByteEditor
from ..documents.emulation_document import EmulationDocument
from ..emulator import guess_emulator

from sawx.utils.processutil import run_detach

import logging
log = logging.getLogger(__name__)


class EmulationEditor(ByteEditor):
    """Editor that holds an emulator instance and the associated viewers.

    Omnivore can theoretically support multiple views of the same emulator
    where each view could show a different restart number/frame number of the
    emulator. But only one should be in control of the operation at any one
    time. So the state of the emulation document would have to be shared, but
    the current frame number would have to reside here in the editor.
    """

    editor_id = "omnivore.emulator"
    ui_name = "Emulator"

    default_viewers = "hex,bitmap,char,disasm"
    default_viewers = "hex,bitmap,disasm"

    preferences_module = "omnivore.editors.emulation_preferences"

    needs_tabless_frame = True

    menubar_desc = [
        ["File",
            ["New",
                "new_blank_file",
                None,
                "new_file_from_template",
            ],
            "open_file",
            ["Open Recent",
                "open_recent",
            ],
            None,
            "save_file",
            "save_as",
            None,
            "save_as_image",
            "save_animation",
            None,
            "quit",
        ],
        ["Edit",
            "undo",
            "redo",
            None,
            "copy",
            "cut",
            "paste",
            None,
            "select_all",
            "select_none",
            "select_invert",
            ["Mark Selection As",
                "disasm_type",
            ],
            None,
            "find",
            "find_expression",
            "find_next",
            "find_to_selection",
            None,
            "prefs",
        ],
        ["View",
            "view_width",
            "view_zoom",
            ["Colors",
                "view_color_standards",
                None,
                "view_antic_powerup_colors",
                None,
                "view_ask_colors",
            ],
            ["Bitmap",
                "view_bitmap_renderers",
            ],
            ["Text",
                "view_font_renderers",
                None,
                "view_font_mappings",
            ],
            ["Font",
                "view_fonts",
                None,
                "view_font_groups",
                None,
                "view_load_font",
                "view_font_from_selection",
                "view_font_from_segment",
            ],
            None,
            ["Add Data Viewer",
                "view_add_viewer",
            ],
            ["Add Emulation Viewer",
                "view_add_emulation_viewer",
            ],
            None,
            "layout_save_emu",
            ["Restore Emulator Layout",
                "layout_restore_emu",
            ],
            None,
            "show_toolbar",
        ],
        ["Bytes",
            "byte_set_to_zero",
            "byte_set_to_ff",
            "byte_nop",
            None,
            "byte_set_high_bit",
            "byte_clear_high_bit",
            "byte_bitwise_not",
            "byte_shift_left",
            "byte_shift_right",
            "byte_rotate_left",
            "byte_rotate_right",
            "byte_reverse_bits",
            "byte_random",
            None,
            "byte_set_value",
            "byte_or_with_value",
            "byte_and_with_value",
            "byte_xor_with_value",
            None,
            "byte_ramp_up",
            "byte_ramp_down",
            "byte_add_value",
            "byte_subtract_value",
            "byte_subtract_from",
            "byte_multiply_by",
            "byte_divide_by",
            "byte_divide_from",
            None,
            "byte_reverse_selection",
            "byte_reverse_group",
        ],
        ["Media",
            "generate_segment_menu()",
            None,
            "comment_add",
            "comment_remove",
            None,
            "segment_goto",
        ],
        ["Machine",
            "emu_pause_resume",
            "emu_prev_history",
            "emu_next_history",
            None,
            "emu_generate_emulator_specific_key_actions()",
            None,
            "emu_warmstart",
            "emu_coldstart",
        ],
        ["Debug",
            "emu_step",
            "emu_step_into",
            "emu_step_over",
            "emu_break_next_scan_line",
            "emu_break_frame",
            "emu_break_vbi_start",
            "emu_break_vbi_end",
            "emu_break_dli_start",
            "emu_break_dli_end",
        ],
        ["Help",
            "about",
            None,
            "user_guide",
            None,
            "show_debug_log",
            "open_log_directory",
            "open_log",
            None,
            ["Developer Debugging",
                "show_focus",
                "widget_inspector",
                "garbage_objects",
                "raise_exception",
                "test_progress",
            ],
        ],
    ]

    keybinding_desc = {
        "emu_boot_disk_image": "Ctrl+T",  # TEST binding
        "emu_a8_start_key": "F4",
        "emu_a8_select_key": "F3",
        "emu_a8_option_key": "F2",
        "emu_warmstart": "F5",
        "emu_coldstart": "Shift-F5",
        "emu_prev_history": "F6",
        "emu_next_history": "F7",
        "emu_pause_resume": "F8",
        "emu_step": "F9",
        "emu_step_over": "F10",
        "emu_step_out": "F11",
        "emu_break_vbi_start": "shift-F9",
        "emu_break_frame": "F12",

    }

    # Convenience functions

    @property
    def emulator(self):
        return self.document.emulator

    def __init__(self, document, *args, **kwargs):
        ByteEditor.__init__(self, document, *args, **kwargs)
        self.document.boot()

    #### ByteEditor interface

    def prepare_destroy(self):
        self.document.stop_timer()
        ByteEditor.prepare_destroy(self)

    def process_keyword_arguments(self, kwargs):
        doc = self.document
        self.selected_checkpoint_range = None
        if kwargs:
            if "skip_frames" in kwargs:
                doc.skip_frames_on_boot = int(kwargs["skip_frames"])

    def preprocess_document(self, source_document, kwargs):
        if source_document.__class__ == EmulationDocument:
            return source_document
        skip = 0
        log.debug(f"preprocess_document: EmulatorEditor, {source_document}, kwargs={kwargs}")
        kwargs = {}
        try:
            source_document.emulator_type
        except:
            try:
                emulator_type = kwargs['machine']
            except KeyError:
                emulator_type = guess_emulator(source_document)
            doc = EmulationDocument.create_document(source_document=source_document, emulator_type=emulator_type)
        log.debug(f"Using emulator {doc.emulator_type}")
        return doc

    #### images (sawx.editor interface)

    def get_numpy_image(self):
        return self.document.emulator.get_frame_rgb()

    #### animation (sawx.editor interface)

    @property
    def numpy_animation_available(self):
        return self.selected_checkpoint_range is not None

    def get_numpy_animation(self):
        print("ANIMATION RANGE:", self.selected_checkpoint_range)
        start, end = self.selected_checkpoint_range
        if start.frame_number > end.frame_number:
            start, end = end, start
        doc = self.document
        emu = doc.emulator
        frames = []
        save = (emu.current_frame_number, emu.current_restart.restart_number)
        restart_number = end.restart_number
        frame_number = end.frame_number
        while frame_number >= start.frame_number:
            emu.restore_restart(restart_number, frame_number)
            frame = emu.get_frame_rgb(frame_number)
            frame = frame.copy()  # need copy rather than reference to same data storage space
            frames.append(frame)
            frame_number -= 1
        frames.reverse()
        emu.restore_restart(*save)
        return (self.document.framerate, frames)

    #### template

    @property
    def layout_template_search_order(self):
        # search for emulator-specific templates before any generic ones
        order = [self.editor_id + "." + self.document.emulator.name, self.editor_id]
        return order

    #### menu

    def emu_generate_emulator_specific_key_actions(self):
        return self.emulator.get_special_key_actions()



from __future__ import division
import cython
import numpy as np
cimport numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
def get_numpy_memory_map_image(segment_viewer, np.ndarray[np.uint8_t, ndim=2] bytes, np.ndarray[np.uint8_t, ndim=2] style, int start_byte, int end_byte, int bytes_per_row, int num_rows, int start_col, int num_cols):
    cdef int num_rows_with_data = (end_byte - start_byte + bytes_per_row - 1) // bytes_per_row
    cdef np.uint8_t bgr = segment_viewer.preferences.background_color[0]
    cdef np.uint8_t bgg = segment_viewer.preferences.background_color[1]
    cdef np.uint8_t bgb = segment_viewer.preferences.background_color[2]
    cdef np.uint8_t sr = segment_viewer.preferences.highlight_background_color[0]
    cdef np.uint8_t sg = segment_viewer.preferences.highlight_background_color[1]
    cdef np.uint8_t sb = segment_viewer.preferences.highlight_background_color[2]
    
    cdef int end_row = min(num_rows_with_data, num_rows)
    cdef int end_col = min(bytes_per_row, start_col + num_cols)
    cdef int width = end_col - start_col
    cdef int height = num_rows_with_data
    cdef np.ndarray[np.uint8_t, ndim=3] array = np.empty([height, width, 3], dtype=np.uint8)

    cdef int y = 0
    cdef int e = start_byte
    cdef int x, i, j
    cdef np.uint8_t bw = 0xff
    cdef np.uint8_t c, s
    cdef np.uint16_t h
    for j in range(end_row):
        x = 0
        for i in range(start_col, end_col):
            if e + i >= end_byte:
                array[y,x,0] = bgr
                array[y,x,1] = bgg
                array[y,x,2] = bgb
            else:
                c = bytes[j, i] ^ bw
                s = style[j, i]
                if s & 0x80:
                    h = sr * c >> 8
                    array[y,x,0] = h
                    h = sg * c >> 8
                    array[y,x,1] = h
                    h = sb * c >> 8
                    array[y,x,2] = h
                else:
                    array[y,x,0] = c
                    array[y,x,1] = c
                    array[y,x,2] = c
            x += 1
        y += 1
        e += bytes_per_row

    return array


# Fast font rendering. As an optimization, only renders complete rectangles.
# The first and last row may be partial depending on the start offset of the
# segment, but these now have to be rendered separately in their own (single
# row) rectangle
@cython.boundscheck(False)
@cython.wraparound(False)
def get_numpy_font_map_image(segment_viewer, antic_font, np.ndarray[np.uint8_t, ndim=2] bytes, np.ndarray[np.uint8_t, ndim=2] style, int start_byte, int end_byte, int bytes_per_row, int num_rows, int start_col, int num_cols):
    cdef int char_w = antic_font.char_w
    cdef int char_h = antic_font.char_h
    cdef int end_col = min(bytes_per_row, start_col + num_cols)
    cdef int width = char_w * num_cols
    cdef int height = num_rows * char_h
    cdef np.ndarray[np.uint8_t, ndim=3] array = np.empty([height, width, 3], dtype=np.uint8)
    cdef np.uint8_t[:,:,:] fast_array = array

    cdef int y = 0
    cdef int x, i, j
    cdef np.ndarray[np.uint8_t, ndim=4] f = antic_font.normal_font
    cdef np.uint8_t[:,:,:,:] fast_f = f
    cdef np.ndarray[np.uint8_t, ndim=4] fh = antic_font.highlight_font
    cdef np.uint8_t[:,:,:,:] fast_fh = fh
    cdef np.ndarray[np.uint8_t, ndim=4] fd = antic_font.data_font
    cdef np.uint8_t[:,:,:,:] fast_fd = fd
    cdef np.ndarray[np.uint8_t, ndim=4] fm = antic_font.match_font
    cdef np.uint8_t[:,:,:,:] fast_fm = fm
    cdef np.ndarray[np.uint8_t, ndim=4] fc = antic_font.comment_font
    cdef np.uint8_t[:,:,:,:] fast_fc = fc
    cdef np.uint8_t s, c
    cdef np.ndarray[np.uint8_t] mapping = segment_viewer.font_mapping.font_mapping
    for j in range(num_rows):
        x = 0
        for i in range(start_col, start_col + num_cols):
            c = mapping[bytes[j, i]]
            s = style[j, i]
            if s & 0x80:
                fast_array[y:y+char_h,x:x+char_w,:] = fast_fh[c]
            elif s & 0x20:
                fast_array[y:y+char_h,x:x+char_w,:] = fast_fm[c]
            elif s & 0x40:
                fast_array[y:y+char_h,x:x+char_w,:] = fast_fc[c]
            elif s & 0x08:
                fast_array[y:y+char_h,x:x+char_w,:] = fast_fd[c]
            else:
                fast_array[y:y+char_h,x:x+char_w,:] = fast_f[c]
            x += char_w
        y += char_h

    return array


@cython.boundscheck(False)
@cython.wraparound(False)
def get_numpy_memory_access_image(segment_viewer, int bytes_per_row, int num_rows, int count, np.ndarray[np.uint8_t, ndim=2] access_value, np.ndarray[np.uint8_t, ndim=2] access_type, int start_col, int num_cols):

    bg = segment_viewer.preferences.highlight_background_color
    cdef np.uint8_t sr = bg[0]
    cdef np.uint8_t sg = bg[1]
    cdef np.uint8_t sb = bg[2]
    
    cdef int end_col = min(bytes_per_row, start_col + num_cols)
    cdef int width = end_col - start_col
    cdef np.ndarray[np.uint8_t, ndim=3] array = np.empty([num_rows, width, 3], dtype=np.uint8)

    cdef int y = 0
    cdef int x, i, j
    cdef np.uint8_t c, s, lo, hi, r, g, b
    cdef np.uint16_t h
    for j in range(num_rows):
        x = 0
        for i in range(start_col, end_col):
            c = access_value[j, i]
            s = access_type[j, i]

            #/* lower 4 bits: bit access flags */
            #/* upper 4 bits: type of access, not a bit field */
            lo = s & 0x0f
            hi = s & 0xf0
            if c > 64 or hi == 0:
                #define ACCESS_TYPE_READ 1
                #define ACCESS_TYPE_WRITE 2
                #define ACCESS_TYPE_EXECUTE 4
                if lo & 0x04:
                    r = 0
                    g = c
                    b = 0
                elif lo & 0x03 == 0x03:
                    r = c
                    g = 0
                    b = c
                elif lo & 0x02:
                    r = c
                    g = 0
                    b = 0
                elif lo & 0x01:
                    r = 0
                    g = 0
                    b = c
                else:
                    r = c
                    g = c
                    b = c
            else:
                #define ACCESS_TYPE_VIDEO 0x10
                #define ACCESS_TYPE_DISPLAY_LIST 0x20
                #define ACCESS_TYPE_CHBASE 0x30
                #define ACCESS_TYPE_PMBASE 0x40
                #define ACCESS_TYPE_CHARACTER 0x50
                #define ACCESS_TYPE_HARDWARE 0x60
                if hi == 0x10:
                    #define ACCESS_TYPE_VIDEO 0x10
                    r = c
                    g = c
                    b = c + 96
                elif hi == 0x20:
                    #define ACCESS_TYPE_DISPLAY_LIST 0x20
                    r = c + 96
                    g = 0
                    b = c + 96
                elif hi == 0x30:
                    #define ACCESS_TYPE_CHBAS 5
                    r = c + 96
                    g = c + 96
                    b = 0
                elif hi == 0x40:
                    #define ACCESS_TYPE_PMBAS 6
                    r = c + 96
                    g = 0
                    b = 0
                elif hi == 0x50:
                    #define ACCESS_TYPE_CHARACTER 8
                    r = c
                    g = c + 96
                    b = c + 96
                elif hi == 0x60:
                    #define ACCESS_TYPE_HARDWARE 0x7f
                    r = c + 96
                    g = c + 96
                    b = c
                else:
                    r = c
                    g = c
                    b = c


            # if s & 0x80:
            #     r = (sr * r) >> 8
            #     g = (sr * g) >> 8
            #     b = (sr * b) >> 8


            array[y,x,0] = r
            array[y,x,1] = g
            array[y,x,2] = b
            x += 1
        y += 1

    return array



# Atari 8-bit color utilities

import math

import numpy as np


# Color references:
#
# Color swatches http://atariage.com/forums/topic/243369-atari-128-color-palettes/


def clamp(val):
    if val < 0.0:
        return 0
    elif val > 255.0:
        return 255
    return int(val)


def ntsc_phase(cr):
    return ((cr-1)*25 - 58) * (2 * math.pi / 360);


def pal_phase(cr):
    return ((cr-1)*25.7 - 15) * (2 * math.pi / 360);


def gtia_to_rgb(val, phasefunc):
    # http://atariage.com/forums/topic/107853-need-the-256-colors/page-2#entry1312467
    cr = (val >> 4) & 15;
    lm = val & 15;
    if cr > 0:
        crlv = 50
    else:
        crlv = 0

    phase = phasefunc(cr)

    y = 255*(lm+1)/16;
    i = crlv*math.cos(phase);
    q = crlv*math.sin(phase);

    r = y + 0.956*i + 0.621*q;
    g = y - 0.272*i - 0.647*q;
    b = y - 1.107*i + 1.704*q;

    return clamp(r), clamp(g), clamp(b)


def gtia_ntsc_to_rgb(val):
    return gtia_to_rgb(val, ntsc_phase)


def gtia_pal_to_rgb(val):
    return gtia_to_rgb(val, pal_phase)


ntsc_iq_lookup = [
    [  0.000,  0.000 ],
    [  0.144, -0.189 ],
    [  0.231, -0.081 ],
    [  0.243,  0.032 ],
    [  0.217,  0.121 ],
    [  0.117,  0.216 ],
    [  0.021,  0.233 ],
    [ -0.066,  0.196 ],
    [ -0.139,  0.134 ],
    [ -0.182,  0.062 ],
    [ -0.175, -0.022 ],
    [ -0.136, -0.100 ],
    [ -0.069, -0.150 ],
    [  0.005, -0.159 ],
    [  0.071, -0.125 ],
    [  0.124, -0.089 ],
    ]


def gtia_ntsc_to_rgb_table(val):
    # This is a better representation of the NTSC colors using a lookup table
    # rather than the phase calculations. Also from the same thread:
    # http://atariage.com/forums/topic/107853-need-the-256-colors/page-2#entry1319398
    cr = (val >> 4) & 15;
    lm = val & 15;

    y = 255*(lm+1)/16;
    i = ntsc_iq_lookup[cr][0] * 255
    q = ntsc_iq_lookup[cr][1] * 255

    r = y + 0.956*i + 0.621*q;
    g = y - 0.272*i - 0.647*q;
    b = y - 1.107*i + 1.704*q;

    return clamp(r), clamp(g), clamp(b)


gtia_ntsc_to_rgb = gtia_ntsc_to_rgb_table


def powerup_colors():
    # Playfield colors are from Mapping the Atari
    # Player/missile colors (the first 4 colors) are normally zero, but I'm
    # specifying others here so they are distinguishable when used.
    return list([4, 30, 68, 213, 40, 202, 148, 70, 0])


def gr0_colors(colors):
    if len(colors) == 5:
        bg_index = 2
        lm_index = 1
    else:
        bg_index = 6
        lm_index = 5
    bg = colors[bg_index]
    cr = bg & 0xf0;
    lm = colors[lm_index] & 0x0f
    fg = cr | lm
    return fg, bg


def calc_playfield_rgb(antic_color_registers, color_converter=None):
    if color_converter is None:
        color_converter = gtia_ntsc_to_rgb
    registers = np.zeros((256, 3), dtype=np.uint8)
    registers[0] = color_converter(antic_color_registers[8])  # COLBG
    registers[1] = color_converter(antic_color_registers[4])  # COLPF0
    registers[2] = color_converter(antic_color_registers[5])  # COLPF1
    registers[3] = color_converter(antic_color_registers[6])  # COLPF2
    registers[4] = color_converter(antic_color_registers[7])  # COLPF3

    # make sure there are 16 registers for 4bpp modes
    for i in range(5, 16):
        registers[i] = (i*16, i*16, i*16)

    # Extend to 32 for dimmed copies of the 16 colors
    for i in range(16, 32):
        r = registers[i - 16]
        registers[i] = (r[0]/4 + 64, r[1]/4 + 64, r[2]/4 + 64)
    return registers


def get_color_registers(antic_color_registers, color_converter):
    registers = []
    for c in antic_color_registers:
        registers.append(color_converter(c))

    # make sure there are 16 registers for 4bpp modes
    i = len(registers)
    for i in range(len(registers), 16):
        registers.append((i*16, i*16, i*16))

    # Extend to 32 for dimmed copies of the 16 colors
    dim = []
    for r in registers:
        dim.append((r[0]/4 + 64, r[1]/4 + 64, r[2]/4 + 64))
    registers.extend(dim)
    return registers


class NTSCConverter:
    name = "NTSC"
    converter = gtia_ntsc_to_rgb

    def __str__(self):
        return self.name

    def __call__(self, colors):
        converter = self.__class__.converter
        return converter(colors)


class PALConverter(NTSCConverter):
    name = "PAL"
    converter = gtia_pal_to_rgb


color_standard_list = [
    NTSCConverter(),
    PALConverter(),
]

valid_color_standards = {item.name: item for item in color_standard_list}



# Don't export the utility functions
__all__ = ['atari_color_to_rgb', 'powerup_colors', 'gr0_colors', 'color_standard_list', 'valid_color_standards']


if __name__ == "__main__":
    def print_opengl_colors(name, func):
        print("%s = np.array([" % name)
        for i in range(256):
            c = func(i)
            c2 = list(c)
            c2.append(255)
            c2 = tuple(c2)
            print(str(c2)+",", end=' ')
        print("])")

    print_opengl_colors("NTSC", gtia_ntsc_to_rgb)
    print_opengl_colors("PAL", gtia_pal_to_rgb)



sha1_signatures = {
"application/x.vectrex": {
  b'J\x83>Hi"\x1f\x83-\xac\xb7\xef\x0f\xb2\x99|\x87\x0b}\x7f': '3-D Crazy Coaster (1983)',
  b'1\x99t`\x9eB\x92\xa0\xe6\xb6\xa7\xc3Bd\xcb\xa0\x82\x85U\t': '3-D Crazy Coaster (1983) [b1]',
  b'\x171\xe8\x92\xda\x89E\xfeIE\xb48G\xd29`\xc5\x81h*': '3-D Mine Storm (1983)',
  b"\x9e]#\xcb~\xf1\x119N\x18\x04\xe1\x05\xb6\xdd'|\x06-\xf6": '3-D Narrow Escape (1983)',
  b'O\xf2\xca\xa1\x1b\xa1\x11\xb2Va\r4\xa1\x9d\xe9{(\x050&': '3-D Narrow Escape (1983) [b1]',
  b'v2\\\xe2\xd6\x96\xdd\xbf\xde\xf7\xc9\xff2\xe0\xb4\xe3\xd3Q\xe1\xe6': '4-D Rotating Cube Demo (19xx) (PD)',
  b'\x85X\x8c\xa1\x16\xfe6\x95\xf1\x9e\x1d\x06\xdd\x08S"\n[/9': 'All Good Things by John Dondzila (1996)',
  b'q\x94k\xe4}\xeb\xae\x89\xad\x19\x1a\xc3\x82\xc2\x01]j\x86 r': 'Animaction (1983) (light pen)',
  b'\xebv\xf8`\x13\x0f\x8c\xc8x&\xfa\x93\x13PFw`Q\x12\x13': 'Armor Attack (1982)',
  b'\x8f5\xba\xc5\xe8\xe0\xfcJ\xd7)\x1b\xbf\xe5\xaeni2r\xde\x06': 'Art Master (1983) (light pen)',
  b'?\xf3\xfaO\xed\xa6\xd5b\xde\xdf\xf2\xa1\x0c\xfa\x86y\xd81\x80D': 'Bedlam (1983)',
  b'\xe4\xba\xddIN\xd5Hr\x13(\xef\x86\x9d3\xadp\xe3\xe5"\xc9': 'Berzerk (1982)',
  b'\x9e"!\xce_u\xb4\x9a\x08L(L\xc9\xdb\r\xf4\xcc\x8ab\xea': 'Berzerk (1982) (Bugfixed Prototype)',
  b'\x88\x03\x18#\xed*\xf9^\x8azb9\xcd\x83\xb1X\xf9\xb7\xc1j': 'Berzerk (1982) (Bugfixed Prototype) [o1]',
  b'\xaa\x1a\xa8\x11\xc9\xce^\x19\xf4;x\xa6S\x1c\xdd\x0eps7\xbd': 'Blitz! - Action Football (1982)',
  b'`\x89=\xfa\xdaB\xcfP\xc9@\x83\xd4Z\xd5\xbd\x14\xd0\xc01\xc8': 'Blitz! - Action Football (1982) [b1]',
  b'g\xf8Q9X\xc0N\x93k\x13W@\xedN\xc6\xe6\xfa\x17c\xd5': 'Clean Sweep (1982)',
  b')\x98A\xa1\x16\xe5\x93\x99\xafNt\xcc\xae#\xdd\xfc\x07\x8e\x00{': 'Clean Sweep (Mr Boston Version) (1982)',
  b'V\xa7\xf2w\xbf\t\xf9Da\xdd\xa5B\xc0L6&m\x83\x8b\x86': 'Cosmic Chasm (1982)',
  b'\xdb\xb0\xb3x\xaa\xcc*\x18\xdf\xb4v\xaa\x99\xab\x00\xd6\xab9R\x05': 'Dark Tower (1983) (Prototype)',
  b'\xfb\xfdXFMB\x98m\x10\xf1\xb3\x7f2\x10q\xectRTi': 'Dark Tower (1983) (Prototype) (Fred Taft Hack) [h1]',
  b"\xeb\x15\xf5\xc9\x0c|1*y\x98\x16\xaa\x9fi\xc6\xfc'\x06>\xce": 'Dark Tower (1983) (Prototype) [b2]',
  b'u\xf9\x1f\xd2\xc1\xefC\xdc\xc8\xf2\x15\xda;\xb9\x92\xfe\x8c!\xf4\x8b': 'Dark Tower (1983) (Prototype) [b3]',
  b'\xbf\x12BY<\x1f\xa0\x8e\x8c6p\xcbnK\x8dQxc\xd0.': 'Dual Vectrex Test #1 (1983) (PD)',
  b'\xc7\xbc\xa2>\x90\xe6\xf8\xad\xe5\xc6h\xf7\xffU\xc6\xc7,*\x08q': 'Dual Vectrex Test #2 (1983) (PD)',
  b'\xc2&\xd7\x83\x01\xdb\xe4\x0f!\x94\xc1e\xf2\x12MZ\xfa\x88[\xe4': 'Engine Analyzer (1983) (light pen)',
  b'(n\x06\x131(\x9b\x03\x1d\xe37%\xa4\x96l7\x9c\x1d\xf4\x7f': 'Fortress of Narzod (1982)',
  b'A \xa6W\xce*p#q\xd0\x9f\xe7\xfdX\xdc\x9c\xc4\xb7\xe5\xcb': 'Gravitrex by John Dondzila (2002)',
  b'\x04\x06\xc19R\xfb\x0b\xd6\xad\xc6\x82\x98\xae\x16>\xbca`HL': 'Heads Up - Action Soccer (1983)',
  b'\xa2F\xf1\xd9J[\x06+`xG\x87\xa1\xd6\xd4\x99\xb2\x1c\x10\xcd': 'Hyperchase (1982)',
  b'X\xae\xcc\x7f\xb1\xd2j"\xb0\x9e\x9a\x13\xd8E\xe4\xed\x12\xe2D\x9d': 'Melody Master (1983) (light pen)',
  b'e\xd0t&\xb5 \xdd\xd3\x11]@\xf2UQ\x1e\x0f\xd2\xe2\n\xe7': 'Mine Storm (1982)',
  b'\xe0\x90\x0b\xe6\xd6\x85\x8b\x98_\xd7\xf0\x99\x9d\x86K/\xce\xaf\x01\xa1': 'Mine Storm (1982) (Karrsoft Hack) [h1]',
  b'\xa9aD\xbb6\xa9N\xb0\xa2\xf9\x81p\x12\x13\xceYy\x11\x85\xd1': 'Mine Storm (1982) (RLB Hack) [h1]',
  b'\xc6\xc4\x0bOV\xd6\xf8L\xb6GX\xd0\ntT\x96\xe3\xe8\x1dh': 'Mine Storm (1982) (RLB Hack) [o1]',
  b'\x11B\xf7W\x0e\xfc\xcd}\x03\xb4\xf5\xa6DrSC\x03\xbf\xcf\xdf': 'Mine Storm (1982) (RLB Hack) [o2]',
  b'\x1f\xdc\xc6\xe5J\xe5\x17{\xc9\xcd\xc7\x9c\xe6\x16\xae4\x01\xe5\xc2)': 'Mine Storm II (1983)',
  b'|\x07R\xc2K6E\xb7\x163{W\xfeoK\xc4\x06N\x9e\xb0': 'Moon Lander Demo by Clay Cowgill (1997) (PD) [a1]',
  b'>\x8f\xfd?V\x9ci\xddJ\xa6G\xaa\x8f9.\xa38\xfe\xfeB': 'Patriots III - BALListics Busters by John Dondzila (2002)',
  b'H\x1a\xa3\xb7a\xf3GG\xfev\x9c\x84 VNO7R\x9c&': 'Patriots by John Dondzila (1996) [a1]',
  b'\x87$\x8d\xcd=\xb5\x1c\x1e\xbf?\x8e\x07\xe2\x1e\xc0!\xee\xc0\xb9\xbe': 'Polar Rescue (1983)',
  b'\xb7\xbd\x8c\x80\x03\xc6\x055\xe1\xe7vu\x8f\\(\x9a\xa9\xa7~\x81': 'Polar Rescue (1983) [a1]',
  b'\xfe\xc1\x13\xdf\xd3\x93\x1f\x14\xa6\xe7\xca\xd2%W\x18`\xf9\xa7"\x1f': 'Polar Rescue (1983) [a1][o1]',
  b'\x84T\xd4rC^zjL\xa5\xfe\x12\x91bH\x0c\xa6\xe8K\x87': 'Pole Position (1982)',
  b'\x9f[\x0b^l\xafm\x88}67\xac\xe7\n\xd6\xa6%\x8b\x99\x9d': 'Pole Position (1982) [f1]',
  b'"\xf7\xa4\x90\xf6\xf7\xaa\xa3\x1e\x89+\x03\xa8Qx\xaf!\x1fb\xe3': 'Rip-Off (1982)',
  b'\xa1\x0f\xfbn\xfd\xf3*c\xf7\xc4\xda\x8b\xb0\xb7\xf8\xd5\xfdl\xdb\x98': 'Rockaroids Remix - 3rd Rock by John Dondzila (1996) [a1]',
  b'8\xe3\x8b\\`FaF\xd4d\x8f\x89)\xb5\xce:\x08\xdc\xbe\r': 'Scramble (1982)',
  b'8\x07-\xc4\x91\xd5\xc7)\xc0oLP\xbc1\r6\x1e\x0f\xf6*': 'Solar Quest (1982)',
  b'\x0c"\x9d\xd1*U\x04F\x81X_\x7f\xda\xbd7\xdb\xc1$;Y': 'Space Wars (1982)',
  b'|\xe8]\xb8\xdd2hz\xd7b\x961\xae\x118 7\x1f\xaf|': 'Spectrum I+ Demo (1982) (PD)',
  b'\x8b\xbag\x05\xe6\xf6\x9fU\xabm\x10^\xac/\xd5\xb0\xc0\xd1m\x9f': 'Spike (1983)',
  b"w]n_\xd48$j%o'\xa2\xa5\x14Sh\xec},\xbe": "Spike Hoppin' by John Dondzila (1998)",
  b'.\x82\x85\x06\x150\xcc\xad\xfa-\x8d\x11\xc3t\xae$DI\xeb\xfa': 'Spinball (1983)',
  b'\x16\x1d\xb6\x8a\\\x87e?\x87\xbb4\xe0\xbd\xd3\x95\xfc\xbf\x0c\xe0y': 'Spinball (1983) [b2]',
  b'\xc9\xb7\x9f\x1b\xa3\r\xcb\xceZ\x06\x8b`j\x97w\\\xf8c\xfe\n': 'Star Castle (1983)',
  b'\xd6\xb3\x88\xb3\x95\x1e4\x97\x1c\x1b\x00\xc7\xfd\xf8\xdd\x00\xadR\xdf\t': 'Star Hawk (1982)',
  b'\xd1\x82\xea\xc5\xa5O\xfcd/\xcb\xe4\xa4\x95\xc7\xc0\x9f\x87\xb45\x85': 'Star Ship (1982)',
  b':.\xe6\x13\x16\xfb1m\x9e\x9e\xfd\xf5\xee@^\x814u\xb4f': 'Star Trek - The Motion Picture (1982)',
  b'!px\x11\xe5D\xed\xaef\xca\x07cQ\xbf\x07C\x84+F\xe5': 'Star Trek - The Motion Picture (1982) [a1]',
  b'\xf3\xd2\xfa,\x1ex\xda\x1faV\x92c\x1e\t\xee\xe1\xe4\xbe\x18\x84': 'Star Trek - The Motion Picture (1982) [b1]',
  b'\xe3A\x80U\xf1\xe7\x9a#\x00W,U~q\xbf,\xe8\xba\xdb\xfa': 'Test Rev. 4 (1982) (PD)',
  b'OF\x1b\x0bQ\xc8$3\xe7\x010\xea\xc5IO&#\x9bDS': 'Tour De France (1983) (Prototype)',
  b'\x7f\x01F\x970\x9e\xfc\xa2/1\x04H\xac\xb0,\x8e]\xa8\x9cj': 'V-Frogger by Chris Salomon (1998) (PD)',
  b'{\xd5\xc2\x96\x86j:q6\xa9\x90\xe1\x1f1&\xd1\x90A\xb3\xce': 'V-Frogger by Chris Salomon (1998) (PD) [a1]',
  b'\xee\xc5t\x1a\xe8Z\xe1\x0eJ\x0b\xa7\x8e\xf7\xee)F\x8b"1@': 'V-Frogger by Chris Salomon, Sound by Kurt Woloch (2001) (PD)',
  b'\x9f`\xae\xf3\xdd\x8a\x8b\x11\x06\xf8Z\x8e\xdfMe:\x0bQW2': 'Vecmania Part 1 (1999) (PD)',
  b'\x93f\xbb\xf1\x7f\xb4J(\xfe\x0e\x1d\xb2\xf4\xc0\xb8\xc3\x1c\xe0\xcf\xe8': 'Vecmania Part 1 (1999) (PD) [a1]',
  b'\x0c\x8b\x89\x9b"\xd3W\x84\xb7>\x1c;=\x91V\xc19\xaa\xc1L': 'Vecmania Part 2 (1999) (PD)',
  b'xYp)3H\x99j\x18\x03\xa8\xd5\x08?\x81\xa0\xf9\xbe\x14\xb9': 'Vectopia by John Dondzila (2001) (PD)',
  b'\x83\xa6&\x9a\x92u\xd8\xceH\xc5\x8bU`\xfb-s$\xea\xee\xa0': 'Vector Vaders by John Dondzila (1996)',
  b'\xb9\xbb\xf5\xbb\x0e\xacR\xd09\xa4\xa9\x93\xa2\xd8\x06K\x86,\x9e(': 'Vectrex BIOS (1982)',
  b'\x96\x81\xdc]8\xa3\x1c\xb0\xd0\xc6\xc3\xbd\x0f\xc7k\xfd\xdeS\x1c\xce': 'Web Warp (1983)',
  b'\x08\xe2\xdf\x9fwi\xc1jMs\n\xdf\x8de\x9f\x8e\xa4\xbe\x7f\x9d': 'Web Wars (1983)',
  b't]\xdb\xdc(_\x1d\xa4\xf7\xc3\xeb^D\xfe%-P<z\x15': 'Wormhole by John Dondzila (2001) (PD)',
},
}  # end sha1_signatures



import os
import sys
import time
import wx
import wx.lib.newevent
try:
    import wx.glcanvas as glcanvas
    import OpenGL.GL as gl
    HAS_OPENGL = True
except ImportError:
    HAS_OPENGL = False

import numpy as np

from sawx.ui.compactgrid import MultiCaretHandler

from ..emulators.atari8bit.colors import NTSC
from .intscale import intscale
from .ui_wx import wxGLSLTextureCanvas, wxLegacyTextureCanvas

import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)


class EmulatorScreenBase(object):
    def __init__(self, emulator):
        self.emulator = emulator

        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.Bind(wx.EVT_KEY_UP, self.on_key_up)

        self.on_size(None)
        if self.IsDoubleBuffered():
            self.Bind(wx.EVT_PAINT, self.on_paint)
        else:
            self.Bind(wx.EVT_PAINT, self.on_paint_double_buffer)

        self.caret_handler = MultiCaretHandler()

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        best = wx.Size(self.emulator.width, self.emulator.height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best

    def on_size(self,evt):
        if not self.IsDoubleBuffered():
            # make new background buffer
            size  = self.GetClientSize()
            self._buffer = wx.Bitmap(size)

    def show_frame(self, frame_number=-1):
        raise NotImplementedError

    def on_key_down(self, evt):
        keycode = evt.GetKeyCode()
        log.debug(f"key down: {keycode}")
        if not self.emulator.process_key_down(evt, keycode):
            evt.Skip()

    def on_key_up(self, evt):
        keycode = evt.GetKeyCode()
        log.debug(f"key up: {keycode}")
        if not self.emulator.process_key_up(evt, keycode):
            evt.Skip()


class BitmapScreen(wx.Panel, EmulatorScreenBase):
    def __init__(self, parent, emulator):
        wx.Panel.__init__(self, parent, -1, size=(emulator.width, emulator.height))
        EmulatorScreenBase.__init__(self, emulator)
        self.scaled_frame = None
        self.image = None
        self.set_scale(1)

    def get_bitmap_slow(self, frame):
        scaled = self.scale_frame(frame)
        h, w, _ = scaled.shape
        image = wx.ImageFromData(w, h, scaled.tostring())
        bmp = wx.BitmapFromImage(image)
        return bmp

    def get_bitmap_fast(self, frame):
        # Slightly improved speed over converting the array to a string
        # slow x 1000: 0.261524
        # fast x 1000: 0.206890
        self.scale_frame(frame)
        # the image data has already been updated because the unterlying
        # numpy array has been changed
        bmp = wx.Bitmap(self.image)
        return bmp

    get_bitmap = get_bitmap_fast

    def bitmap_benchmark(self):
        import time
        t0 = time.clock()
        for i in range(1000):
            self.get_bitmap_slow(frame)
        print("slow x 1000: %f" % (time.clock() - t0))
        t0 = time.clock()
        for i in range(1000):
            self.get_bitmap_fast(frame)
        print("fast x 1000: %f" % (time.clock() - t0))

    def set_scale(self, scale):
        """Scale a numpy array by an integer factor

        This turns out to be too slow to be used by the screen display. OpenGL
        displays don't use this at all because the display hardware scales
        automatically.
        """
        self.screen_scale = scale
        h = self.emulator.height
        w = self.emulator.width
        if scale > 1:
            self.scaled_frame = np.empty((h * scale, w * scale, 3), dtype=np.uint8)
            self.image = wx.ImageFromBuffer(w * scale, h * scale, self.scaled_frame)
        else:
            self.scaled_frame = None
            self.image = wx.ImageFromBuffer(w, h, self.emulator.screen_rgb)

        # self.delay = 5 * scale * scale
        # self.stop_timer()
        # self.start_timer(True)

    def scale_frame(self, frame):
        if self.screen_scale == 1:
            return frame
        scaled = intscale(frame, self.screen_scale, self.scaled_frame)
        log.debug("panel scale: %d, %s" % (self.screen_scale, scaled.shape))
        return scaled

    def show_frame(self, frame_number=-1, force=False):
        if force or frame_number >= 0:
            dc = wx.ClientDC(self)
            self.updateDrawing(dc, frame_number)
        else:
            #self.updateDrawing()
            self.Refresh()

    def updateDrawing(self, dc, frame_number=-1):
        #dc=wx.BufferedDC(wx.ClientDC(self), self._buffer)
        frame = self.emulator.get_frame_rgb(frame_number)
        bmp = self.get_bitmap(frame)
        dc.DrawBitmap(bmp, 0,0, True)

    def on_paint(self, evt):
        dc=wx.PaintDC(self)
        self.updateDrawing(dc)
        self.refreshed=True

    def on_paint_double_buffer(self, evt):
        dc=wx.BufferedPaintDC(self,self._buffer)
        self.updateDrawing(dc)
        self.refreshed=True


class OpenGLEmulatorMixin(object):
    def get_raw_texture_data(self, frame_number=-1):
        raw = self.emulator.get_frame_rgba_opengl(frame_number)
        return raw

    def show_frame(self, frame_number=-1):
        if not self.finished_init:
            return
        frame = self.get_rgba_texture_data(frame_number)
        try:
            self.update_texture(self.display_texture, frame)
        except Exception as e:
            import traceback

            print(traceback.format_exc())
            sys.exit()
        self.on_draw()

    def on_paint(self, evt):
        if not self.finished_init:
            self.init_context()
        self.show_frame()

    on_paint_double_buffer = on_paint


class OpenGLScreen(OpenGLEmulatorMixin, wxLegacyTextureCanvas, EmulatorScreenBase):
    def __init__(self, parent, emulator):
        wxLegacyTextureCanvas.__init__(self, parent, NTSC, -1, size=(3*emulator.width, 3*emulator.height))
        EmulatorScreenBase.__init__(self, emulator)

    def get_rgba_texture_data(self, frame_number=-1):
        raw = self.emulator.get_frame_rgba_opengl(frame_number)
        log.debug("raw data for legacy version: %s" % str(raw.shape))
        return raw


class GLSLScreen(OpenGLEmulatorMixin, wxGLSLTextureCanvas, EmulatorScreenBase):
    def __init__(self, parent, emulator):
        wxGLSLTextureCanvas.__init__(self, parent, NTSC, -1, size=(3*emulator.width, 3*emulator.height))
        EmulatorScreenBase.__init__(self, emulator)

    def get_color_indexed_texture_data(self, frame_number=-1):
        raw = np.flipud(self.emulator.get_color_indexed_screen(frame_number))
        log.debug("raw data for GLSL version: %s" % str(raw.shape))
        return raw



import os
import glob

import numpy as np

from . import cputables
from .flags import *
from . import dtypes as ud
from . import libudis


HistoryStorage = libudis.HistoryStorage
StringifiedHistory = libudis.StringifiedHistory



import os
import sys

import numpy as np

import wx

from atrip import Container, Segment

from sawx.ui import compactgrid as cg
from ..editors.linked_base import VirtualTableLinkedBase

from ..ui.segment_grid import SegmentGridControl

from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)



class SampleVirtualTable(cg.VirtualTable):
    column_labels = ["Target", "Count", "Flag", "Type", "^Values"]
    column_sizes = [4, 2, 2, 2, 40]

    def __init__(self, linked_base):
        self.virtual_linked_base = linked_base
        s = linked_base.segment
        cg.VirtualTable.__init__(self, len(self.column_labels), s.origin)

    def calc_num_rows(self):
        num_cols = self.items_per_row
        return (self.virtual_linked_base.document_length + num_cols - 1) // num_cols

    def calc_last_valid_index(self):
        return self.virtual_linked_base.document_length

    def get_label_at_index(self, index):
        return str(index)

    def get_value_style(self, row, col):
        text = f"r{row}c{col}"
        s = self.virtual_linked_base.segment
        index, _ = self.get_index_range(row, col)
        style = s.style[index]
        return text, style

    def rebuild(self):
        v = self.virtual_linked_base
        old_size = v.document_length
        size = old_size + 1
        old_segment = v.segment
        container = Container(np.arange(size, dtype=np.uint8))
        segment = Segment(container)
        segment.style[0:old_size] = old_segment.style[0:old_size]
        v.segment = segment
        self.init_boundaries()
        print(f"new size: {len(v.segment)}")


class SampleGridControl(SegmentGridControl):
    default_table_cls = SampleVirtualTable

    def calc_default_table(self, linked_base):
        return self.default_table_cls(linked_base)

    def calc_line_renderer(self):
        return cg.VirtualTableLineRenderer(self, 2, widths=self.default_table_cls.column_sizes, col_labels=self.default_table_cls.column_labels)

    def recalc_view(self):
        self.table.rebuild()
        cg.CompactGrid.recalc_view(self)

    def refresh_view(self):
        if self.IsShown():
            log.debug("refreshing %s" % self)
            self.recalc_view()
        else:
            log.debug("skipping refresh of hidden %s" % self)


class VirtualTestViewer(SegmentViewer):
    name = "vtest"

    ui_name = "Virtual Test Viewer"

    control_cls = SampleGridControl

    # initialization

    @classmethod
    def replace_linked_base(cls, linked_base):
        # the new linked base decouples the cursor here from the other segments
        segment = DefaultSegment(np.arange(400, dtype=np.uint8))
        return VirtualTableLinkedBase(editor=linked_base.editor, segment=segment)

    def create_post(self):
        self.linked_base.table = self.control.table

    # properties

    @property
    def table(self):
        return self.control.table

    def byte_values_changed(self, index_range):
        log.debug("byte_values_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.table.rebuild()

    def byte_style_changed(self, index_range):
        log.debug("byte_style_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.table.rebuild()

    def recalc_data_model(self):
        self.table.rebuild()

    def do_priority_level_refresh(self):
        self.control.recalc_view()
        self.refresh_view(True)



import os
import sys

import wx
import numpy as np

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg

from ..ui.segment_grid import SegmentGridControl, SegmentVirtualTable
from ..viewer import SegmentViewer
from .emulator import EmulatorViewerMixin
from . import bitmap as b
from ..arch.bitmap_renderers import MemoryAccessMap

import logging
log = logging.getLogger(__name__)


class MemoryAccessTable(SegmentVirtualTable):
    segment_name = "memaccess"
    segment_ui_name = "Memory Access"

    def get_data_style_view(self, linked_base):
        self.access_segment = linked_base.document.find_segment_by_name("Memory Access")
        self.type_segment = linked_base.document.find_segment_by_name("Access Type")
        data = self.access_segment.data
        style = self.type_segment.data
        return data, style

    def calc_num_cols(self):
        return 256

    def get_label_at_index(self, index):
        return(str(hex(index)))


class MemoryAccessGridControl(b.BitmapGridControl):
    default_table_cls = MemoryAccessTable

    def set_viewer_defaults(self):
        self.items_per_row = 256
        self.zoom = 1

    @property
    def bitmap_renderer(self):
        return MemoryAccessMap()

    @property
    def pixels_per_byte(self):
        return 1

    @property
    def scale_width(self):
        return self.bitmap_renderer.scale_width

    @property
    def scale_height(self):
        return self.bitmap_renderer.scale_height


class MemoryAccessViewer(EmulatorViewerMixin, SegmentViewer):
    name = "mem"

    ui_name = "Memory Access"

    viewer_category = "Emulator"

    control_cls = MemoryAccessGridControl

    has_bitmap = True

    has_colors = True

    has_width = False

    has_zoom = True

    zoom_text = "bitmap zoom factor"

    priority_refresh_frame_count = 1

    def update_bitmap(self, evt):
        log.debug("BitmapViewer: machine bitmap changed for %s" % self.control)
        if evt is not Undefined:
            self.control.recalc_view()
            self.linked_base.editor.update_pane_names()

    def validate_width(self, width):
        return 256

    def recalc_data_model(self):
        self.control.recalc_view()

    @property
    def current_frame_number(self):
        return self.linked_base.emulator.current_frame_number



# Standard library imports.
import sys
import os
import uuid

# Major package imports.
import wx
import numpy as np
import json

from atrip import Container, ContainerHeader, Segment

from sawx.utils.command import DisplayFlags
from sawx.events import EventHandler

from ..commands import CoalescingChangeByteCommand
from ..utils.segmentutil import Segment

import logging
log = logging.getLogger(__name__)


blank_data = np.zeros([1], dtype=np.uint8)
blank_container = Container(blank_data)
blank_segment = Segment(blank_container)

class LinkedBase:
    """Model for the state of a set of viewers. A ByteEditor can have an
    arbitrary number of LinkedBases, but all viewers that point to a common
    LinkedBase will all show the same data.

    A LinkedBase is tied to a Segment, and any data that can be applied
    segment-wide should also live here, e.g. the disassembly trace.

    Currently, a LinkedBase contains a single Machine, which specifies how to
    look at the data, like: CPU used for disassembly/assembly, available
    graphics and text modes, current colors and font. All views of the data
    using this LinkedBase will show the same Machine.

    In the future, each view will have its own Machine, because there's nothing
    in the Machine that could affect other viewers, so it should really be part
    of the view.
    """
    rect_select = False

    def __init__(self, editor, segment=None):
        self.uuid = str(uuid.uuid4())
        self.editor = editor

        if segment is None:
            segment = blank_segment
        self.segment = segment
        self.segment_uuid = segment.uuid
        self.has_origin = False
        self.segment_view_params = {}
        self.current_selection = None

        self.ensure_visible_event = EventHandler(self)
        self.sync_caret_event = EventHandler(self)
        self.sync_caret_to_index_event = EventHandler(self)
        self.recalc_event = EventHandler(self)
        self.update_trace = EventHandler(self)
        self.key_pressed = EventHandler(self)
        self.segment_selected_event = EventHandler(self)

        ##### Jumpman-specific stuff
        self.jumpman_trigger_selected_event = EventHandler(self)

    #### Properties

    @property
    def section_name(self):
        return str(self.segment)

    @property
    def window(self):
        return self.editor.window

    @property
    def frame(self):
        return self.editor.frame

    @property
    def document(self):
        return self.editor.document

    @property
    def document_length(self):
        return len(self.segment)

    @property
    def emulator(self):
        return self.editor.emulator

    @property
    def cached_preferences(self):
        return self.editor.preferences

    ##### Cleanup

    def prepare_for_destroy(self):
        self.editor = None

    #### Convenience functions

    def __str__(self):
        return f"LinkedBase {hex(id(self))}: seg={self.segment}"

    def restore_session(self, e):
        log.debug("metadata: %s" % str(e))
        if 'uuid' in e:
            self.uuid = e['uuid']
        if 'diff highlight' in e:
            self.diff_highlight = bool(e['diff highlight'])
        if 'segment view params' in e:
            self.segment_view_params = e['segment view params']
        if 'segment_uuid' in e:
            self.view_segment_uuid(e['segment_uuid'], False)

    def serialize_session(self, mdict):
        # make sure to save latest values of currently viewed segment
        self.save_segment_view_params(self.segment)

        mdict['uuid'] = self.uuid
        mdict['segment view params'] = self.segment_view_params
        mdict['segment_uuid'] = self.segment_uuid

    def save_segment_view_params(self, segment):
        d = {}
        for viewer in self.editor.viewers:
            if viewer.linked_base == self:
                try:
                    d[viewer.uuid] = viewer.control.calc_view_params()
                except AttributeError:
                    pass
        log.debug("segment view params: %s: %s" % (segment.name, str(d)))
        self.segment_view_params[segment.uuid] = d

    def restore_segment_view_params(self, segment):
        try:
            d = self.segment_view_params[segment.uuid]
        except KeyError:
            log.debug("no view params for %s" % segment.uuid)
            d = {}
        else:
            log.debug("restoring view params for segment %s (%s): %s" % (segment.name, segment.uuid, str(d)))
        for viewer in self.editor.viewers:
            if viewer.linked_base == self:
                try:
                    params = d[viewer.uuid]
                except KeyError:
                    viewer.use_default_view_params()
                    continue
                try:
                    log.debug(" restoring view of %s (%s): %s" % (viewer.window_title, viewer.uuid, str(params)))
                    viewer.restore_view_params(params)
                except AttributeError:
                    continue

    def find_first_valid_segment_uuid(self):
        s = self.document.collection.find_boot_media()
        return s.uuid

    def find_segment(self, segment, refresh=False, data_model_changed=True):
        if segment is not None:
            uuid = segment.uuid
        elif self.segment_uuid == None:
            uuid = self.find_first_valid_segment_uuid()
        else:
            uuid = self.segment_uuid
        log.debug("find_segment: changing from %d to %d, input=%s" % (self.segment_uuid, uuid, segment))

        if refresh:
            self.view_segment_uuid(uuid)
            data_model_changed = False
        else:
            self.segment_uuid = uuid
            self.segment = self.document.collection[uuid]

        if data_model_changed:
            self.force_data_model_update()
            self.restore_segment_view_params(self.segment)
            self.task.segments_changed = self.document.segments  # FIXME
            self.segment_selected_event(self.segment_uuid)

    def view_segment_uuid(self, uuid, do_callbacks=True):
        log.debug(f"view_segment_uuid: changing to {uuid} from {self.segment_uuid}")
        doc = self.document
        if uuid is None:
            uuid = self.find_first_valid_segment_uuid()
        if uuid != self.segment_uuid:
            old_segment = self.segment
            if old_segment is not None and old_segment != blank_segment:
                self.save_segment_view_params(old_segment)
            self.segment = doc.collection[uuid]
            self.segment_uuid = uuid
            if do_callbacks:
                self.recalc_event(True)
                self.adjust_selection(old_segment)

                #self.show_trace()
                self.segment_selected_event(self.segment_uuid)
                #self.task.status_bar.message = "Switched to segment %s" % str(self.segment)
                #self.task.update_window_title()
        log.debug(f"view_segment_uuid: changed to {self.segment_uuid}")

    def force_refresh(self):
        flags = DisplayFlags()
        flags.byte_values_changed = True
        self.process_flags(flags)

    def sync_caret_to_index(self, index):
        flags = DisplayFlags()
        flags.caret_index = index
        print(f"sync_caret_to_index: index={index}")
        self.sync_caret_to_index_event(flags=flags)

    #### command flag processors

    def process_flags(self, flags):
        self.editor.process_flags(flags)

    #### convenience functions

    def change_bytes(self, start, end, byte_values, pretty=None):
        """Convenience function to perform a ChangeBytesCommand
        """
        self.document.change_count += 1
        cmd = CoalescingChangeByteCommand(self.segment, start, end, byte_values)
        if pretty:
            cmd.ui_name = pretty
        self.editor.process_command(cmd)

    # #### CaretHandler overrides

    # def calc_caret_history(self):
    #     return self.segment, CaretHandler.calc_caret_state(self)

    # def restore_caret_history(self, state):
    #     segment, carets = state
    #     number = self.document.find_segment_index(segment)
    #     if number < 0:
    #         log.error("tried to restore caret to a deleted segment? %s" % segment)
    #     else:
    #         if number != self.segment_uuid:
    #             self.view_segment_uuid(number)
    #         CaretHandler.restore_caret_state(self, carets)
    #     log.debug(self.caret_history)

    # def collapse_selections_to_carets(self):
    #     CaretHandler.collapse_selections_to_carets(self)
    #     self.document.change_count += 1
    #     self.segment.clear_style_bits(selected=True)

    #### selection utilities

    @property
    def can_copy(self):
        return self.current_selection is not None

    def set_current_selection(self, caret_handler):
        log.debug(f"set_current_selection: {caret_handler}")
        self.current_selection = caret_handler
        return caret_handler.get_selected_status_message()

    def clear_current_selection(self):
        log.debug(f"clear_current_selection")
        self.current_selection = None

    def adjust_selection(self, old_segment):
        """Adjust the selection of the current segment so that it is limited to the
        bounds of the new segment.
        
        If the current selection is entirely out of bounds of the new segment,
        all the selection indexes will be set to zero.
        """
        # find byte index of view into master array
        g = old_segment.container
        s = self.segment

        self.restore_segment_view_params(s)
        if False:  # select same bytes in new segment, if possible
            self.selected_ranges = s.get_style_ranges(selected=True)
            if self.selected_ranges:
                # Arbitrarily puth the anchor on the last selected range
                last = self.selected_ranges[-1]
                self.anchor_initial_start_index = self.anchor_start_index = last[0]
                self.anchor_initial_end_index = self.anchor_end_index = last[1]
            g.clear_style_bits(selected=True)
        self.document.change_count += 1

    def get_segments_from_selection(self, size=-1):
        s = self.segment
        segments = []

        # Get the selected ranges directly from the segment style data, because
        # the selected ranges in the caret list can be out of order or
        # overlapping
        ranges = s.get_style_ranges(selected=True)
        if len(ranges) == 1:
            seg_start, seg_end = ranges[0]
            if size < 0:
                size = seg_end - seg_start
            for start in range(seg_start, seg_end, size):
                end = min(seg_end, start + size)
                segment = DefaultSegment(s.rawdata[start:end], s.origin + start)
                segments.append(segment)
        elif len(ranges) > 1:
            # If there are multiple selections, use an indexed segment
            indexes = []
            for start, end in ranges:
                indexes.extend(list(range(start, end)))
            if size < 0:
                size = len(indexes)
            for i in range(0, len(indexes), size):
                raw = s.rawdata.get_indexed(indexes[i:i + size])
                segment = DefaultSegment(raw, s.origin + indexes[i])
                segments.append(segment)
        return segments

    def index_clicked(self, index, bit, from_control, refresh_from=True):
        log.debug("index_clicked: %s from %s at %d, %s" % (refresh_from, from_control, index, bit))
        self.carets.current.index = index
        if refresh_from:
            from_control = None
        self.update_caret = (from_control, index, bit)


class VirtualTableLinkedBase(LinkedBase):
    """A LinkedBase for virtual data; that is, data that is computed on the fly
    or data that is generated and doesn't physically reside in the segment
    data. It is linked to a HexTable that provides the sizing information.
    """

    def __init__(self, *args, **kwargs):
        LinkedBase.__init__(self, *args, **kwargs)
        self.table = None

    @property
    def document_length(self):
        return self.table.last_valid_index + 1

    def find_first_valid_segment_uuid(self):
        log.debug("segment change operations ignored on VirtualLinkedBase")

    def find_segment(self, segment, refresh=False, data_model_changed=True):
        log.debug("segment change operations ignored on VirtualLinkedBase")

    def view_segment_uuid(self, uuid):
        log.debug("segment change operations ignored on VirtualLinkedBase")



sha1_signatures = {
"application/x.atari2600.starpath": {
  b'L\xda\x82\x17l\xab\xc3\xe52\xeb\xca\xd0\xf63\x17\x9al@;\xae': 'Escape from the Mindmaster (Labyrinth) (1982) (Starpath Corporation, Dennis Caswell) (5) (AR-4200) (PAL)',
  b'\xad\xb2B\x98E~\x15\xfa\xac\xdaS-P\xb7%;\xef\x0e\xf3\xb2': 'Escape from the Mindmaster (Labyrinth) (1982) (Starpath Corporation, Dennis Caswell) (5) (AR-4200) ~',
},
}  # end sha1_signatures



import sys
import pytest

import numpy as np

from atrip import Container, Segment
from atrip.memory_map import MemoryMap
from atrip.disassembler import ParsedDisassembly, DisassemblyConfig, dd


class TestNOP:
    cpu_id = 10  # 6502
    data = np.zeros(256, dtype=np.uint8) + 0xea
    expected_count = 256
    expected = [
        b'nop',
        b'nop',
        b'nop',
        b'nop',
    ]
    labels = None

    def setup(self):
        c = Container(self.data)
        segment = Segment(c, origin=0x6000)
        driver = DisassemblyConfig()
        segment.style[:] = 0
        segment.disasm_type[:] = self.cpu_id
        self.parsed = driver.parse(segment, 8000)
        self.entries = self.parsed.entries
        if self.__class__.labels is not None:
            self.labels = MemoryMap.from_list("test labels", self.labels)
        print("raw data: ", type(self.data), self.data)
        print("data", segment.data, segment.data.tobytes())
        print("style", segment.style, segment.style.tobytes())
        print("disasm_type", segment.disasm_type, segment.disasm_type.tobytes())

    def test_disasm(self):
        p = self.parsed
        e = self.entries
        assert len(p) == self.expected_count
        # for i in range(len(p)):
        #     print(e[i])
        # print(p.index_to_row)

        t = p.stringify(0, 100, labels=self.labels)
        # for line in t:
        #     print(line)
        for i, text in enumerate(self.expected):
            assert t[i] == text


class TestORA(TestNOP):
    data = np.asarray([1,1,1,0x80,1,2,1,0xff, 0,0,0,0,0,0,0,0, 0,1,1,1,1,1,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0], dtype=np.uint8)
    expected_count = 33
    expected = [
        b'ora (L0001,x)',
        b'ora (L0080,x)',
        b'ora (L0002,x)',
        b'ora (L00ff,x)',
        b'brk',
    ]


class TestLabels(TestORA):
    expected = [
        b'ora (L0001,x)',
        b'ora (ADDR80,x)',
        b'ora (L0002,x)',
        b'ora (ADDRFF,x)',
        b'brk',
    ]
    labels = [(0x80, "ADDR80"), (0xff, "ADDRFF")]


def sample():
    if len(sys.argv) > 1:
        with open(sys.argv[1], 'rb') as fh:
            data = np.frombuffer(fh.read(), dtype=np.uint8)
            data = data.astype(dtype=np.uint8)  # prevents data.base from being a 'bytes' object
    else:
        data = repeat_test
    print("raw data: ", type(data), data)
    c = Container(data)
    segment = Segment(c, origin=0x6000)
    print("data", segment.data, segment.data.tobytes())
    print("style", segment.style, segment.style.tobytes())
    print("disasm_type", segment.disasm_type, segment.disasm_type.tobytes())

    if False:
        p = ParsedDisassembly(1000, 0x6000)
        print(p)
        p.parse_test("6502", data)
        print(p.entries)
        e = p.entries.view(dtype=dd.HISTORY_ENTRY_DTYPE)
        # for i in range(len(e)):
        #     print(e[i])
        # print(f"number of entries: {len(e)}")
        # print(hex(len(p.labels)))
        # used_labels = np.where(p.labels != 0)[0]
        # print(used_labels)
        # for index in used_labels:
        #     print(index, p.labels[index])

        num_lines = 100
        for i in range(num_lines):
            print(e[i])
        p.stringify(0, num_lines, False, True)
        for i in range(num_lines):
            start = p.text_starts[i]
            count = p.line_lengths[i]
            text = p.text_buffer[start:start + count].tostring()
            print(f"text[{start}:{start + count}] = {text}, {e[i]}")

    driver = DisassemblyConfig()
    segment.style[:] = 0
    segment.disasm_type[:] = 10
    p = driver.parse(segment, 8000)
    e = p.entries
    print(p)
    print(len(p))
    for i in range(len(p)):
        print(e[i])
    print(p.index_to_row)

    t = p.stringify(0,100)
    print(t)
    for line in t:
        print(line)

    del p
    print("deleted")

if __name__ == "__main__":
    # t = TestNOP()
    # t = TestORA()
    t = TestLabels()
    t.setup()
    t.test_disasm()



import numpy as np

atascii_to_internal = np.hstack([np.arange(64, 96, dtype=np.uint8),np.arange(64, dtype=np.uint8),np.arange(96, 128, dtype=np.uint8)])
atascii_to_internal = np.hstack([atascii_to_internal, atascii_to_internal + 128])

internal_to_atascii = np.hstack([np.arange(32, 96, dtype=np.uint8),np.arange(32, dtype=np.uint8),np.arange(96, 128, dtype=np.uint8)])
internal_to_atascii = np.hstack([internal_to_atascii, internal_to_atascii + 128])



class ATASCIIFontMapping(object):
    name = "ASCII Order"
    font_mapping = atascii_to_internal

    def __str__(self):
        return self.name

    def __eq__(self, other):
        try:
            return self.name == other.name
        except AttributeError:
            pass
        return False

    # to be usable in dicts, py3 needs __hash__ defined if __eq__ is defined
    def __hash__(self):
        return id(self)

    def wx_char_to_byte(self, char, mods, control):
        import wx
        byte = None

        if mods == wx.MOD_RAW_CONTROL:
            if char == 44:  # Ctrl-, prints ATASCII 0 (heart)
                byte = 0 + control.inverse
            elif char >= 65 and char <= 90:  # Ctrl-[A-Z] prints ATASCII chars 1-26
                byte = char - 64 + control.inverse
            elif char == 46:  # Ctrl-. prints ATASCII 96 (diamond)
                byte = 96 + control.inverse
            elif char == 59:  # Ctrl-; prints ATASCII 123 (spade)
                byte = 123 + control.inverse
            elif char == wx.WXK_TAB:
                byte = 158
            elif char == 50:  # Ctrl-2 prints ATASCII 253 (buzzer)
                byte = 253
            elif char == wx.WXK_INSERT:
                byte = 255
        elif mods == wx.MOD_SHIFT:
            if char == wx.WXK_BACK:
                byte = 156
            elif char == wx.WXK_INSERT:
                byte = 157
            elif char == wx.WXK_TAB:
                byte = 159
        elif char == wx.WXK_HOME:
            byte = 125
        elif char == wx.WXK_BACK:
            byte = 126
        elif char == wx.WXK_TAB:
            byte = 127
        elif char == wx.WXK_RETURN:
            byte = 155
        elif char == wx.WXK_DELETE:
            byte = 254
        elif char == wx.WXK_INSERT:
            byte = 255

        elif control.pending_esc:
            if char == wx.WXK_ESCAPE:
                byte = 27
            elif char == wx.WXK_UP:
                byte = 28
            elif char == wx.WXK_DOWN:
                byte = 29
            elif char == wx.WXK_LEFT:
                byte = 30
            elif char == wx.WXK_RIGHT:
                byte = 31

        elif char == wx.WXK_ESCAPE:
            control.pending_esc = True

        return byte

    def convert_byte_mapping(self, char):
        return char


class AnticFontMapping(ATASCIIFontMapping):
    name = "Antic Order"
    font_mapping = np.arange(256, dtype=np.uint8)

    def convert_byte_mapping(self, char):
        try:
            char = atascii_to_internal[char]
        except IndexError:
            pass
        return char


font_mapping_list = [
    ATASCIIFontMapping(),
    AnticFontMapping(),
]

valid_font_mappings = {mapping.name: mapping for mapping in font_mapping_list}



import os
import sys
import zlib
import json

import logging
log = logging.getLogger(__name__)

from ._version import __version__

try:
    import numpy as np
except ImportError:
    raise RuntimeError("atrip %s requires numpy" % __version__)

from . import errors
from .container import Container, ContainerHeader, guess_container
from .collection import Collection
from .segment import Segment
from . import style_bits
# from .ataridos import AtrHeader, AtariDosDiskImage, BootDiskImage, AtariDosFile, XexContainerSegment, get_xex, add_atr_header
# from .dos33 import Dos33DiskImage
# from .segments import SegmentData, SegmentSaver, DefaultSegment, EmptySegment, ObjSegment, RawSectorsSegment, SegmentedFileSegment, interleave_segments, SegmentList
# from .style_bits import user_bit_mask, match_bit_mask, comment_bit_mask, data_style, selected_bit_mask, diff_bit_mask, not_user_bit_mask, get_style_mask, get_style_bits
# from .spartados import SpartaDosDiskImage
# from .cartridge import A8CartHeader, AtariCartImage, RomImage
# from .parser import SegmentParser, DefaultSegmentParser, guess_parser_by_size, guess_parser_for_mime, guess_parser_for_system, iter_parsers, iter_known_segment_parsers, mime_parse_order, parsers_for_filename
# from .container import guess_container
# from .magic import guess_detail_for_mime
# from .utils import to_numpy, text_to_int
# from .dummy import LocalFilesystem


def process(image, dirent, options):
    skip = False
    action = "copying to"
    filename = dirent.get_filename()
    outfilename = filename
    if options.no_sys:
        if dirent.ext == "SYS":
            skip = True
            action = "skipping system file"
    if not skip:
        if options.xex:
            outfilename = "%s%s.XEX" % (dirent.filename, dirent.ext)
    if options.lower:
        outfilename = outfilename.lower()

    if options.dry_run:
        action = "DRY_RUN: %s" % action
        skip = True
    if options.extract:
        print("%s: %s %s" % (dirent, action, outfilename))
        if not skip:
            bytes = image.get_file(dirent)
            with open(outfilename, "wb") as fh:
                fh.write(bytes)
    else:
        print(dirent)


def find_diskimage_from_data(data, verbose=False):
    data = to_numpy(data)
    parser = None
    container = guess_container(data, verbose)
    if container is not None:
        data = container.unpacked
    rawdata = SegmentData(data)
    mime, parser = guess_parser_by_size(rawdata)
    if parser is None:
        for mime in mime_parse_order:
            if verbose:
                print("Trying MIME type %s" % mime)
            parser = guess_parser_for_mime(mime, rawdata, verbose)
            if parser is None:
                continue
            if verbose:
                print("Found parser %s" % parser.menu_name)
            mime2 = guess_detail_for_mime(mime, rawdata, parser)
            if mime != mime2:
                mime = mime2
                if verbose:
                    print("Magic signature match: %s" % mime)
            break
    if parser is None:
        raise errors.UnsupportedDiskImage("Unknown disk image type")
    return parser, mime


def find_container_from_data(name, data, verbose=False):
    container = guess_container(data)
    container.name = name
    container.guess_media_type()
    container.guess_filesystem()
    return container


def find_container(filename, verbose=False):
    sample_data = np.fromfile(filename, dtype=np.uint8)
    return find_container_from_data(os.path.basename(filename), sample_data, verbose)


def find_collection(filename, verbose=False):
    sample_data = np.fromfile(filename, dtype=np.uint8)
    if len(sample_data) > 0:
        log.info(f"loaded {len(sample_data)} bytes from {filename}")
    else:
        raise IOError("No data")
    return Collection(filename, sample_data)


def extract_files(collection, files):
    files = set(files)
    for dirent in collection.iter_dirents():
        if not files or dirent.filename in files:
            output = dirent.filename
            if options.lower:
                output = output.lower()
            if not options.dry_run:
                segment = dirent.get_file()
                if os.path.exists(output) and not options.force:
                    print("skipping %s, file exists. Use -f to overwrite" % output)
                else:
                    print("extracting %s -> %s" % (dirent.filename, output))
                    with open(output, "wb") as fh:
                        fh.write(segment[:])
            else:
                print("extracting %s -> %s" % (dirent.filename, output))
            if dirent.filename in files:
                files.remove(dirent.filename)
                if not files:
                    break
    if files:
        for filename in sorted(files):
            print(f"not found: {filename}")


def save_file(image, name, filetype, data):
    try:
        dirent = image.find_dirent(name)
        if options.force:
            image.delete_file(name)
        else:
            print("skipping %s, use -f to overwrite" % (name))
            return False
    except errors.FileNotFound:
        pass
    print("copying %s to %s" % (name, image.filename))
    if not options.dry_run:
        image.write_file(name, filetype, data)
        return True
    return False


def add_files(image, files):
    filetype = options.filetype
    if not filetype:
        filetype = image.default_filetype
    changed = False
    for name in files:
        with open(name, "rb") as fh:
            data = fh.read()
        changed = save_file(image, name, filetype, data)
    if changed:
        image.save()


def remove_files(image, files):
    changed = False
    for name in files:
        try:
            dirent = image.find_dirent(name)
        except errors.FileNotFound:
            print("%s not in %s" % (name, image))
            continue
        print("removing %s from %s" % (name, image))
        if not options.dry_run:
            image.delete_file(name)
            changed = True
    if changed:
        image.save()


def list_files(collection, files, show_crc=False, show_metadata=False):
    files = set(files)
    for dirent in collection.iter_dirents():
        if not files or dirent.filename in files:
            if show_crc:
                data = image.get_file(dirent)
                crc = zlib.crc32(data) & 0xffffffff  # correct for some platforms that return signed int
                extra = "  %08x" % crc
            else:
                extra = ""
            print("%s%s" % (dirent, extra))
            if show_metadata:
                print(dirent.extra_metadata(image))


def crc_files(image, files):
    files = set(files)
    for dirent in image.files:
        if not files or dirent.filename in files:
            data = image.get_file(dirent)
            crc = zlib.crc32(data) & 0xffffffff  # correct for some platforms that return signed int
            print("%s: %08x" % (dirent.filename, crc))


def assemble_segments(source_files, data_files, obj_files, run_addr=""):
    if source_files:
        try:
            import pyatasm
        except ImportError:
            raise errors.AtrError("Please install pyatasm to compile code.")
    changed = False
    segments = SegmentList()
    for name in source_files:
        try:
            asm = pyatasm.Assemble(name)
        except SyntaxError as e:
            raise errors.AtrError("Assembly error: %s" % e.msg)
        log.info("Assembled %s into:" % name)
        for first, last, object_code in asm.segments:
            s = segments.add_segment(object_code, first)
            log.info("  %s" % s.name)
            print("adding %s from %s assembly" % (s, name))
    for name in data_files:
        if "@" not in name:
            raise errors.AtrError("Data files must include a load address specified with the @ char")
        name, addr = name.rsplit("@", 1)
        first = text_to_int(addr)
        log.info("Adding data file %s at $%04x" % (name, first))
        subset = slice(0, sys.maxsize)
        if "[" in name and "]" in name:
            name, slicetext = name.rsplit("[", 1)
            if ":" in slicetext:
                start, end = slicetext.split(":", 1)
                try:
                    start = int(start)
                except:
                    start = 0
                if end.endswith("]"):
                    end = end[:-1]
                try:
                    end = int(end)
                except:
                    end = None
                subset = slice(start, end)
        with open(name, 'rb') as fh:
            data = fh.read()[subset]
            s = segments.add_segment(data, first)
            log.debug("read data for %s" % s.name)
    for name in obj_files:
        try:
            parser, _ = find_diskimage(name)
        except errors.AtrError as e:
            log.warning(f"skipping {name}: {e}")
        else:
            for s in parser.segments:
                if hasattr(s, 'run_address'):
                    if not run_addr:
                        run_addr = s.run_address()
                    else:
                        log.warning(f"already have run address {run_addr}; skipping {s.run_address()}")
                elif s.origin > 0:
                    log.info(f"adding {s} from {name}")
                    segments.add_segment(s.data, s.origin)
    for s in segments:
        log.info("(%s - %04x)" % (str(s)[:-1], s.origin + len(s)))
    if run_addr:
        try:
            run_addr = text_to_int(run_addr)
        except (AttributeError, ValueError):
            # not text, try as integer
            try:
                run_addr = int(run_addr)
            except ValueError:
                run_addr = None

    return segments, run_addr

def assemble(image, source_files, data_files, obj_files, run_addr=""):
    segments, run_addr = assemble_segments(source_files, data_files, obj_files, run_addr)
    file_data, filetype = image.create_executable_file_image(options.output, segments, run_addr)
    print("total file size: $%x (%d) bytes" % (len(file_data), len(file_data)))
    changed = save_file(image, options.output, filetype, file_data)
    if changed:
        image.save()


def boot_image(image_name, source_files, data_files, obj_files, run_addr=""):
    try:
        image_cls = parsers_for_filename(image_name)[0]
    except errors.InvalidDiskImage as e:
        log.warning("%s: %s" % (image_name, e))
        return None
    segments, run_addr = assemble_segments(source_files, data_files, obj_files, run_addr)
    if segments:
        image = image_cls.create_boot_image(segments, run_addr)
        print("saving boot disk %s" % (image_name))
        image.save(image_name)
    else:
        print("No segments to save to boot disk")


def shred_image(image, value=0):
    print("shredding: free sectors from %s filled with %d" % (image, value))
    if not options.dry_run:
        image.shred()
        image.save()


def get_template_path(rel_path="templates"):
    path = __file__

    template_path = os.path.normpath(os.path.join(os.path.dirname(path), rel_path))
    frozen = getattr(sys, 'frozen', False)
    if frozen:
        if frozen == True:
            # pyinstaller sets frozen=True and uses sys._MEIPASS
            root = sys._MEIPASS
            template_path = os.path.normpath(os.path.join(root, template_path))
        elif frozen  == 'macosx_app':
            #print "FROZEN!!! %s" % frozen
            root = os.environ['RESOURCEPATH']
            if ".zip/" in template_path:
                zippath, template_path = template_path.split(".zip/")
            template_path = os.path.normpath(os.path.join(root, template_path))
        else:
            print("App packager %s not yet supported for image paths!!!")
    return template_path


def get_template_images(partial=""):
    import glob

    path = get_template_path()
    files = glob.glob(os.path.join(path, "*"))
    templates = {}
    for path in files:
        name = os.path.basename(path)
        if name.endswith(".inf"):
            continue
        if partial not in name:
            continue
        try:
            with open(path + ".inf", "r") as fh:
                s = fh.read()
                try:
                    j = json.loads(s)
                except ValueError:
                    continue
                j['name'] = name
                j['path'] = path
                templates[name] = j
        except IOError:
            continue
    return templates


def get_template_info():
    import textwrap
    fmt = "  %-14s  %s"

    templates = get_template_images()

    lines = []
    lines.append("available templates:")
    for name in sorted(templates.keys()):
        d = textwrap.wrap(templates[name]["description"], 80 - 1 - 14 - 2 - 2)
        lines.append(fmt % (os.path.basename(name), d[0]))
        lines.extend([fmt % ("", line) for line in d[1:]])
    return os.linesep.join(lines) + os.linesep


def get_template_data(template):
    possibilities = get_template_images(template)
    if not possibilities:
        raise errors.InvalidDiskImage("Unknown template disk image %s" % template)
    if len(possibilities) > 1:
        raise errors.InvalidDiskImage("Name %s is ambiguous (%d matches: %s)" % (template, len(possibilities), ", ".join(sorted(possibilities.keys()))))
    name, inf = possibilities.popitem()
    path = inf['path']
    try:
        with open(path, "rb") as fh:
            data = fh.read()
    except IOError:
        raise errors.InvalidDiskImage("Failed reading template file %s" % path)
    return data, inf


def create_image(template, name):
    import textwrap

    try:
        data, inf = get_template_data(template)
    except errors.InvalidDiskImage as e:
        info = get_template_info()
        print("Error: %s\n\n%s" % (e, info))
        return
    print("Using template %s:\n  %s" % (inf['name'], "\n  ".join(textwrap.wrap(inf["description"], 77))))
    if not options.dry_run:
        if os.path.exists(name) and not options.force:
            print("skipping %s, use -f to overwrite" % (name))
        else:
            with open(name, "wb") as fh:
                fh.write(data)
            parser, _ = find_diskimage(name)
            print("created %s: %s" % (name, str(parser.image)))
            list_files(parser.image, [])
    else:
        print("creating %s" % name)


def trace_calls(frame, event, arg):
    # global last_trace_was_system_call

    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    if "/python3" in caller_filename or "/logging/" in func_filename or "/sre_" in caller_filename or "/logging/" in caller_filename:
        # if not last_trace_was_system_call:
        #     print('  <system calls>')
        #     last_trace_was_system_call = True
        return
    last_trace_was_system_call = False
    print('%s:%s -> %s %s:%s' % (caller_filename, caller_line_no, func_name, func_filename, func_line_no))
    return


def run():
    import argparse

    global options

    # Subparser command aliasing from: https://gist.github.com/sampsyo/471779
    # released into the public domain by its author
    class AliasedSubParsersAction(argparse._SubParsersAction):
        class _AliasedPseudoAction(argparse.Action):
            def __init__(self, name, aliases, help):
                dest = name
                if aliases:
                    dest += ' (%s)' % ','.join(aliases)
                sup = super(AliasedSubParsersAction._AliasedPseudoAction, self)
                sup.__init__(option_strings=[], dest=dest, help=help) 

        def add_parser(self, name, **kwargs):
            if 'aliases' in kwargs:
                aliases = kwargs['aliases']
                del kwargs['aliases']
            else:
                aliases = []

            parser = super(AliasedSubParsersAction, self).add_parser(name, **kwargs)

            # Make the aliases work.
            for alias in aliases:
                self._name_parser_map[alias] = parser
            # Make the help text reflect them, first removing old help entry.
            if 'help' in kwargs:
                help = kwargs.pop('help')
                self._choices_actions.pop()
                pseudo_action = self._AliasedPseudoAction(name, aliases, help)
                self._choices_actions.append(pseudo_action)

            return parser

    command_aliases = {
        "list": ["t", "ls", "dir", "catalog"],
        "crc": [],
        "extract": ["x"],
        "add": ["a"],
        "create": ["c"],
        "boot": ["b"],
        "assemble": ["s", "asm"],
        "delete": ["rm", "del"],
        "vtoc": ["v"],
        "segments": [],
        "menu": [],
    }
    # reverse aliases does the inverse mapping of command aliases, including
    # the identity mapping of "command" to "command"
    reverse_aliases = {z: k for k, v in command_aliases.items() for z in (v + [k])}

    skip_diskimage_summary = set(["crc"])

    usage = "%(prog)s [-h] [-v] [--dry-run] COMMAND [args] DISK_IMAGE [...]"
    subparser_usage = " [-h] [-v] [--dry-run] DISK_IMAGE"

    parser = argparse.ArgumentParser(prog="atrip COMMAND DISK_IMAGE", description="Manipulate files on several types of 8-bit computer disk images. Type '%(prog)s COMMAND --help' for list of options available for each command.")
    # parser.register('action', 'parsers', AliasedSubParsersAction)
    parser.add_argument("-v", "--verbose", default=0, action="count")
    parser.add_argument("--dry-run", action="store_true", default=False, help="don't perform operation, just show what would have happened")
    parser.add_argument("--trace", action="store_true", default=False, help="trace program flow (debugging only)")

    subparsers = parser.add_subparsers(dest='command', help='', metavar="COMMAND")
    subparsers.default = "list"

    command = "list"
    p = subparsers.add_parser(command, help="List files on the disk image. This is the default if no command is specified", aliases=command_aliases[command])
    p.add_argument("-g", "--segments", action="store_true", default=False, help="display segments")
    p.add_argument("-m", "--metadata", action="store_true", default=False, help="show extra metadata for named files")
    p.add_argument("-c", "--crc", action="store_true", default=False, help="compute CRC32 for each file")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")
    p.add_argument("files", metavar="FILENAME", nargs="*", help="an optional list of files to display")

    command = "crc"
    p = subparsers.add_parser(command, help="List files on the disk image and the CRC32 value in format suitable for parsing", aliases=command_aliases[command])
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")
    p.add_argument("files", metavar="FILENAME", nargs="*", help="an optional list of files to display")

    command = "extract"
    p = subparsers.add_parser(command, help="Copy files from the disk image to the local filesystem", aliases=command_aliases[command])
    p.add_argument("-a", "--all", action="store_true", default=False, help="operate on all files on disk image")
    p.add_argument("-l", "--lower", action="store_true", default=False, help="convert extracted filenames to lower case")
    #p.add_argument("-n", "--no-sys", action="store_true", default=False, help="only extract things that look like games (no DOS or .SYS files)")
    p.add_argument("-e", "--ext", action="store", nargs=1, default=False, help="add the specified extension")
    p.add_argument("-f", "--force", action="store_true", default=False, help="allow file overwrites on local filesystem")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")
    p.add_argument("files", metavar="FILENAME", nargs="*", help="if not using the -a/--all option, a file (or list of files) to extract from the disk image.")

    command = "add"
    p = subparsers.add_parser(command, help="Add files to the disk image", aliases=command_aliases[command])
    p.add_argument("-f", "--force", action="store_true", default=False, help="allow file overwrites in the disk image")
    p.add_argument("-t", "--filetype", action="store", default="", help="file type metadata for writing to disk images that require it (e.g. DOS 3.3)")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")
    p.add_argument("files", metavar="FILENAME", nargs="+", help="a file (or list of files) to copy to the disk image")

    command = "create"
    p = subparsers.add_parser(command, help="Create a new disk image", aliases=command_aliases[command], epilog="<generated on demand to list available templates>", formatter_class=argparse.RawDescriptionHelpFormatter)
    p.add_argument("-f", "--force", action="store_true", default=False, help="replace disk image file if it exists")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")
    p.add_argument("template", metavar="TEMPLATE", nargs=1, help="template to use to create new disk image; see below for list of available built-in templates")

    command = "assemble"
    p = subparsers.add_parser(command, help="Create a new binary file in the disk image", aliases=command_aliases[command])
    p.add_argument("-f", "--force", action="store_true", default=False, help="allow file overwrites in the disk image")
    p.add_argument("-s", "--asm", nargs="*", action="append", help="source file(s) to assemble using pyatasm")
    p.add_argument("-d","--data", nargs="*", action="append", help="binary data file(s) to add to assembly, specify as file@addr. Only a portion of the file may be included; specify the subset using standard python slice notation: file[subset]@addr")
    p.add_argument("-b", "--obj", "--bload", nargs="*", action="append", help="binary file(s) to add to assembly, either executables or labeled memory dumps (e.g. BSAVE on Apple ][), parsing each file's binary segments to add to the resulting disk image at the load address for each segment")
    p.add_argument("-r", "--run-addr", "--brun", action="store", default="", help="run address of binary file if not the first byte of the first segment")
    p.add_argument("-o", "--output", action="store", default="", required=True, help="output file name in disk image")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")

    command = "boot"
    p = subparsers.add_parser(command, help="Create a bootable disk image", aliases=command_aliases[command])
    p.add_argument("-f", "--force", action="store_true", default=False, help="allow file overwrites in the disk image")
    p.add_argument("-s", "--asm", nargs="*", action="append", help="source file(s) to assemble using pyatasm")
    p.add_argument("-d","--data", nargs="*", action="append", help="binary data file(s) to add to assembly, specify as file@addr. Only a portion of the file may be included; specify the subset using standard python slice notation: file[subset]@addr")
    p.add_argument("-b", "--obj", "--bload", nargs="*", action="append", help="binary file(s) to add to assembly, either executables or labeled memory dumps (e.g. BSAVE on Apple ][), parsing each file's binary segments to add to the resulting disk image at the load address for each segment")
    p.add_argument("-r", "--run-addr", "--brun", action="store", default="", help="run address of binary file if not the first byte of the first segment")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")

    command = "delete"
    p = subparsers.add_parser(command, help="Delete files from the disk image", aliases=command_aliases[command])
    p.add_argument("-f", "--force", action="store_true", default=False, help="remove the file even if it is write protected ('locked' in Atari DOS 2 terms), if write-protect is supported disk image")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")
    p.add_argument("files", metavar="FILENAME", nargs="+", help="a file (or list of files) to remove from the disk image")

    command = "vtoc"
    p = subparsers.add_parser(command, help="Show a formatted display of sectors free in the disk image", aliases=command_aliases[command])
    p.add_argument("-e", "--clear-empty", action="store_true", default=False, help="fill empty sectors with 0")
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")

    command = "segments"
    p = subparsers.add_parser(command, help="Show the list of parsed segments in the disk image", aliases=command_aliases[command])
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")

    command = "menu"
    p = subparsers.add_parser(command, help="Show the segment hierarchy in the disk image", aliases=command_aliases[command])
    p.add_argument("disk_image", metavar="DISK_IMAGE", nargs=1, help="disk image")


    # argparse doesn't seem to allow a default command, so if the first
    # argument isn't recognized, use the "list" command
    args = list(sys.argv[1:])
    if len(args) > 0:
        for i, arg in enumerate(args):
            if arg.startswith("-"):
                continue
            else:
                if arg not in reverse_aliases:
                    args[i:i] = ["list"]
                break
    else:
        parser.print_help()
        sys.exit(1)

    options = parser.parse_args(args)
    command = reverse_aliases[options.command]

    # Turn off debug messages by default
    level = logging.WARNING
    if options.verbose >= 1:
        level = logging.INFO
    if options.verbose >= 2:
        level = logging.DEBUG
    logging.basicConfig(level=level)
    log = logging.getLogger("atrip")

    log.debug("command line arguments: {args}")

    if options.trace:
        sys.settrace(trace_calls)

    disk_image_name = options.disk_image[0]

    if command == "create":
        create_image(options.template[0], disk_image_name)
    elif command == "boot":
        asm = options.asm[0] if options.asm else []
        data = options.data[0] if options.data else []
        obj = options.obj[0] if options.obj else []
        boot_image(disk_image_name, asm, data, obj, options.run_addr)
    else:
        try:
            collection = find_collection(disk_image_name)
        except (errors.UnsupportedAlgorithm, errors.UnsupportedDiskImage, IOError) as e:
            print(f"{disk_image_name}: {e}")
        else:
            if command not in skip_diskimage_summary:
                print(f"{disk_image_name}: {collection}")
            if command == "vtoc":
                vtoc = container.get_vtoc_object()
                print(vtoc)
                if options.clear_empty:
                    shred_image(container)
            elif command == "list":
                list_files(collection, options.files, options.crc, options.metadata)
            elif command == "crc":
                crc_files(container, options.files)
            elif command == "add":
                add_files(container, options.files)
            elif command == "delete":
                remove_files(container, options.files)
            elif command == "extract":
                extract_files(collection, options.files)
            elif command == "assemble":
                asm = options.asm[0] if options.asm else []
                data = options.data[0] if options.data else []
                obj = options.obj[0] if options.obj else []
                assemble(container, asm, data, obj, options.run_addr)
            elif command == "segments":
                print(collection.verbose_info)
            elif command == "menu":
                for item, level in collection.iter_menu():
                    print(f"{level}: {item}")



"""Segment actions
"""
import os
import sys

import wx

from sawx.action import SawxAction, SawxRadioListAction

from .. import commands
from .. import errors

import logging
log = logging.getLogger(__name__)


class SegmentEntry:
    spaces_per_level = 4

    def __init__(self, segment, level):
        self.segment = segment
        self.indent_level = level * self.spaces_per_level

    def __str__(self):
        return (" " * self.indent_level) + str(self.segment)


class segment_select(SawxRadioListAction):
    empty_list_name = "No Valid Segments"

    def is_valid_segment(self, segment):
        return True

    def calc_list_items(self):
        valid_segments = []
        doc = self.editor.document
        if doc is not None and doc.collection is not None:
            log.debug(f"segment_select: prefix={self.prefix}, list={self.action_list_id}")
            if self.action_list_id:
                container_index = int(self.action_list_id)
                container = doc.collection.containers[container_index]
                it = container.iter_menu(0)
            else:
                it = doc.collection.iter_menu()
            for segment, level in it:
                if self.is_valid_segment(segment):
                    valid_segments.append(SegmentEntry(segment, level))
        return valid_segments

    def calc_checked_list_item(self, action_key, index, item):
        collection = self.editor.document.collection
        try:
            segment = collection[item.segment.uuid]
        except KeyError:
            raise errors.RecreateDynamicMenuBar
        return self.editor.segment_uuid == item.segment.uuid

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.editor.view_segment_uuid(item.segment.uuid)


def prompt_for_comment(e, s, ranges, desc):
    existing = s.get_first_comment(ranges)
    text = prompt_for_string(e.window.control, desc, "Add Comment", existing)
    if text is not None:
        cmd = SetCommentCommand(s, ranges, text)
        e.process_command(cmd)


class AddCommentAction(SawxAction):
    """Add a text comment to a byte location.

    A comment is associated with a single byte, so although a range can be
    selected, the comment is applied to only the first byte in the range.

    Bytes with comments will be highlighted in all displays.
    """
    name = 'Add Comment'
    accelerator = 'Alt+C'

    def is_range(self, event):
        return self.active_editor.can_copy

    def get_index(self, event):
        return self.viewer.linked_base.carets.current.index

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        if self.is_range(event):
            ranges = s.get_style_ranges(selected=True)
            if len(ranges) == 1:
                desc = "Enter comment for first byte of range:\n%s" % e.get_label_of_first_byte(ranges)
            else:
                desc = "Enter comment for first byte of each range:\n%s" % e.get_label_of_first_byte(ranges)
        else:
            ranges = []
        if not ranges:
            index = self.get_index(event)
            ranges = [(index, index+1)]
            desc = "Enter comment for location %s" % index
        prompt_for_comment(e, s, ranges, desc)


class AddCommentPopupAction(AddCommentAction):
    name = 'Add Comment'

    def is_range(self, event):
        return event.popup_data["in_selection"]

    def get_index(self, event):
        return event.popup_data["index"]


class RemoveCommentAction(SawxAction):
    """Remove any comments that are in the selected range, or if no selection
    from the current caret position.

    """
    name = 'Remove Comment'
    accelerator = 'Shift+Alt+C'

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        if e.can_copy:
            ranges = s.get_style_ranges(selected=True)
        else:
            index = e.caret_index
            ranges = [(index, index+1)]
        if ranges:
            cmd = ClearCommentCommand(s, ranges)
            e.process_command(cmd)


class RemoveCommentPopupAction(SawxAction):
    name = 'Remove Comment'

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        if event.popup_data["in_selection"]:
            ranges = s.get_style_ranges(selected=True)
        else:
            index = event.popup_data["index"]
            ranges = [(index, index+1)]
        if ranges:
            cmd = ClearCommentCommand(s, ranges)
            e.process_command(cmd)


def prompt_for_label(e, s, addr, desc):
    existing = s.memory_map.get(addr, "")
    text = prompt_for_string(e.window.control, desc, "Add Label", existing)
    if text is not None:
        cmd = SetLabelCommand(s, addr, text)
        e.process_command(cmd)


class AddLabelAction(SawxAction):
    """Add a label to a byte location.

    Like `Add Comment`_, a label is associated with a single byte, so although
    a range can be selected, the comment is applied to only the first byte in
    the range.

    Unlike comments, labels are *not* highlighted and are only shown in the
    disassembly window.
    """
    name = 'Add Label'
    accelerator = 'Alt+L'
    enabled_name = 'has_origin'

    def get_ranges(self, editor, segment, event):
        if editor.can_copy:  # has selected ranges
            ranges = segment.get_style_ranges(selected=True)
        else:
            ranges = [(editor.caret_index, editor.caret_index + 1)]
        return ranges

    def process_ranges(self, editor, segment, ranges):
        index = ranges[0][0]
        addr = index + segment.origin
        return addr

    def process(self, editor, segment, addr):
        desc = "Enter label for address $%04x" % addr
        prompt_for_label(editor, segment, addr, desc)

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        ranges = self.get_ranges(e, s, event)
        addr = self.process_ranges(e, s, ranges)
        self.process(e, s, addr)


class AddLabelPopupAction(AddLabelAction):
    def get_ranges(self, editor, segment, event):
        if event.popup_data["in_selection"]:
            ranges = segment.get_style_ranges(selected=True)
        else:
            index = event.popup_data["index"]
            ranges = [(index, index+1)]
        return ranges


class RemoveLabelAction(AddLabelAction):
    """Remove the label at the current caret position, or if there is a
    selection, all labels in the selected range.

    """
    name = 'Remove Label'
    accelerator = 'Shift+Alt+L'

    def process_ranges(self, editor, segment, ranges):
        return ranges

    def process(self, editor, segment, ranges):
        cmd = ClearLabelCommand(segment, ranges)
        editor.process_command(cmd)


class RemoveLabelPopupAction(RemoveLabelAction):
    def get_ranges(self, editor, segment, event):
        if event.popup_data["in_selection"]:
            ranges = segment.get_style_ranges(selected=True)
        else:
            index = event.popup_data["index"]
            ranges = [(index, index+1)]
        return ranges



"""Byte modification actions
"""
import os
import sys

import wx

from sawx.action import SawxAction, SawxRadioListAction

from .. import commands
from .. import errors
from atrip.disassembler import valid_cpu_ids, cpu_id_to_name, processors
from ..utils.templateutil import calc_available_memory_maps, available_memory_maps

import logging
log = logging.getLogger(__name__)


class doc_cpu(SawxRadioListAction):
    def calc_name(self, action_key):
        cpu_name = self.get_item(action_key)
        return cpu_name

    def calc_list_items(self):
        return [cpu_id_to_name[i] for i in valid_cpu_ids]

    def calc_checked_list_item(self, action_key, index, item):
        return self.editor.document.cpu == item

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.editor.document.cpu = item

class doc_os_labels(SawxRadioListAction):
    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item)

    def calc_list_items(self):
        items = sorted(calc_available_memory_maps().values())
        return items

    def calc_checked_list_item(self, action_key, index, item):
        return self.editor.document.operating_system == item.keyword

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.editor.document.operating_system = item.keyword



import hashlib
import inspect
import pkg_resources

import numpy as np

from . import errors
from . import style_bits
from .segment import Segment
from .utils import to_numpy, to_numpy_list, uuid
from . import filesystem

import logging
log = logging.getLogger(__name__)


class FileType(Segment):
    """Base class for a file contained in a filesystem.
    """
    ui_name = "Unknown file type"

    extra_serializable_attributes = []

    def __init__(self, media, name, offset, length=0):
        self.media = media
        Segment.__init__(self, media, offset, name=name, length=length)
        self.segments = self.calc_segments()
        self.style_segments()

    def __str__(self):
        s = Segment.__str__(self) + " " + self.ui_name
        return s

    def calc_segments(self):
        """Calculate the segments composing this file, or raise a FileError if
        the segment is incompatible with the file structure represented by this
        class.
        """
        return []

    def style_segments(self):
        """Add styling info to segments.

        This happens only after the successful completion of calc_segments. If
        styling were done in calc_segments, FileTypes that were attempted but
        raised errors would add styling info that would be invalid for the
        file type eventually determined to be correct.
        """
        pass


ignore_base_class_file_types = set([FileType])

_file_types = None

def _find_file_types():
    file_types = []
    for entry_point in pkg_resources.iter_entry_points('atrip.file_types'):
        mod = entry_point.load()
        log.debug(f"find_file_type: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and FileType in obj.__mro__[1:] and obj not in ignore_base_class_file_types:
                log.debug(f"find_file_types:   found file_type class {name}")
                file_types.append(obj)
    return file_types

def find_file_types():
    global _file_types

    if _file_types is None:
        _file_types = _find_file_types()
    return _file_types

def guess_file_type(media, filename, offset, length=0):
    for m in find_file_types():
        log.debug(f"trying file_type {m.ui_name}")
        try:
            found = m(media, filename, offset, length)
        except errors.FileError as e:
            log.debug(f"found error: {e}")
            continue
        else:
            log.info(f"found file_type {m.ui_name} for {filename}")
            return found
    log.info(f"No recognized file type for {filename}")
    return FileType(media, filename, offset, length)



from __future__ import print_function
from builtins import zip
from builtins import range
from builtins import object
import os

import numpy as np
import pytest

from atrip.container import Container
from atrip.segment import Segment
import atrip.style_bits as style_bits
import atrip.errors as errors
from atrip.utils import collapse_values, restore_values
from functools import reduce



class TestSegment:
    def setup(self):
        data = np.arange(4096, dtype=np.uint8)
        data[1::2] = np.repeat(np.arange(16, dtype=np.uint8), 128)
        data[::100] = 0x7f
        self.container = Container(data)
        self.segment = Segment(self.container)
        index_by_100 = np.arange(40, dtype=np.int32) * 100
        self.seg100 = Segment(self.segment, index_by_100)
        self.seg1000 = Segment(self.seg100, [0,10,20,30])

    def samples(self):
        assert self.segment[100] == 0x7f
        assert self.container.data[100] == self.segment[100]
        assert self.segment[100] == self.seg100[1]

        self.segment[100] = 0xff
        assert self.segment[100] == 0xff
        assert self.container.data[100] == self.segment[100]
        assert self.segment[100] == self.seg100[1]

    def test_slice(self):
        self.segment[10:30] = 55
        assert self.segment[20] == 55

        self.seg100[0:11] = np.arange(200, 211)
        assert self.seg100[0] == 200
        assert self.seg1000[0] == 200
        assert self.segment[0] == 200

        assert self.seg100[1] == 201
        assert self.segment[100] == 201

        assert self.seg100[2] == 202
        assert self.segment[200] == 202

        assert self.seg100[10] == 210
        assert self.seg1000[1] == 210
        assert self.segment[1000] == 210

    def test_reverse_offset(self):
        """Test mapping from container element to element in Segment"""
        r1000 = self.seg1000.reverse_offset
        assert r1000[0] == 0
        assert np.all(r1000[1:1000] == -1)
        assert r1000[1000] == 1
        assert r1000[2000] == 2
        assert r1000[3000] == 3

    def test_comments(self):
        s = self.segment
        s100 = self.seg100
        s1000 = self.seg1000
        s.set_comment_at(4, "test4")
        s.set_comment_at(100, "test100")
        s.set_comment_at(400, "test400")
        s.set_comment_at(1000, "test1000")

        assert s.get_comment_at(4) == "test4"
        assert s.style[4] & style_bits.comment_bit_mask
        assert s.get_comment_at(100) == "test100"
        assert s.style[100] & style_bits.comment_bit_mask
        assert s.get_comment_at(400) == "test400"
        assert s.style[400] & style_bits.comment_bit_mask
        assert s.get_comment_at(1000) == "test1000"
        assert s.style[1000] & style_bits.comment_bit_mask

        assert s100.get_comment_at(1) == "test100"
        assert s100.style[1] & style_bits.comment_bit_mask
        assert s100.get_comment_at(4) == "test400"
        assert s100.style[4] & style_bits.comment_bit_mask
        assert s100.get_comment_at(10) == "test1000"
        assert s100.style[10] & style_bits.comment_bit_mask
        assert s1000.get_comment_at(1) == "test1000"
        assert s1000.style[1] & style_bits.comment_bit_mask

        s100.set_comment_at(4, "new400")
        s100.set_comment_at(10, "new1000")
        s1000.set_comment_at(2, "new2000")

        assert s.get_comment_at(400) == "new400"
        assert s.get_comment_at(1000) == "new1000"
        assert s.get_comment_at(2000) == "new2000"
        assert s100.get_comment_at(10) == "new1000"
        assert s100.get_comment_at(20) == "new2000"

        s.set_comment_at(150, "test150")
        s100.clear_comment_ranges([[1,5]])
        assert s.get_comment_at(100) == ""
        assert s.style[100] & style_bits.comment_bit_mask == 0
        assert s.get_comment_at(150) == "test150"
        assert s.style[150] & style_bits.comment_bit_mask
        assert s.get_comment_at(400) == ""
        assert s.style[400] & style_bits.comment_bit_mask == 0
        assert s.get_comment_at(1000) == "new1000"
        assert s.style[1000] & style_bits.comment_bit_mask

    def test_style(self):
        s = self.segment
        s100 = self.seg100
        s1000 = self.seg1000
        s.set_style_ranges([[200, 4000]], selected=True)
        s.set_style_ranges([[1200, 3000]], match=True)
        assert s.style[0] == 0
        assert s100.style[1] == 0
        assert s100.style[10] == style_bits.selected_bit_mask
        assert s100.style[20] == style_bits.selected_bit_mask | style_bits.match_bit_mask
        assert s1000.style[0] == 0
        assert s1000.style[1] == s100.style[10]
        assert s1000.style[2] == s100.style[20]

    def test_metadata(self):
        s = self.segment
        s.set_style_ranges([[200, 400]], selected=True)
        s.set_comment_at(190, "test190")
        s.set_comment_at(210, "test210")
        s.set_comment_at(230, "test230")
        indexes = np.arange(180,220)
        m = s.calc_selected_index_metadata(indexes)
        print(m)
        e = s.encode_selected_index_metadata(*m)

        r = s.restore_selected_index_metadata(e)
        print(r)
        assert r[0].tolist() == m[0].tolist()
        assert r[1].tolist() == m[1].tolist()
        assert r[2] == m[2]

    def test_disasm_type(self):
        s = self.segment
        s100 = self.seg100
        s1000 = self.seg1000
        s.set_disasm_ranges([[200, 4000]], 2)
        s.set_disasm_ranges([[1200, 3000]], 50)
        assert s.disasm_type[0] == 128
        assert s100.disasm_type[1] == 128
        assert s100.disasm_type[10] == 2
        assert s100.disasm_type[20] == 50
        assert s1000.disasm_type[0] == 128
        assert s1000.disasm_type[1] == s100.disasm_type[10]
        assert s1000.disasm_type[2] == s100.disasm_type[20]

        c = s.container
        assert c.disasm_type[2000] == 50

        ranges = collapse_values(c.disasm_type)
        assert ranges == [[128, 0, 200], [2, 200, 1200], [50, 1200, 3000], [2, 3000, 4000], [128, 4000, 4096]]

        c.disasm_type[:] = 0
        assert c.disasm_type[2000] == 0

        restore_values(c.disasm_type, ranges)
        assert s.disasm_type[0] == 128
        assert s100.disasm_type[1] == 128
        assert s100.disasm_type[10] == 2
        assert s100.disasm_type[20] == 50
        assert s1000.disasm_type[0] == 128
        assert s1000.disasm_type[1] == s100.disasm_type[10]
        assert s1000.disasm_type[2] == s100.disasm_type[20]
        assert c.disasm_type[2000] == 50

        r2 = collapse_values(c.disasm_type)
        assert ranges == r2

#         s2 = self.sub_segment
#         print(len(s2))
#         copy = s2.get_comment_locations()
#         print(copy)
#         # comments at 4 and 40 in the original means 2 and 38 in the copy
#         orig = s.get_comment_locations()
#         assert copy[2] == orig[4]
#         assert copy[28] == orig[38]

#     def test_split_data_at_comment(self):
#         s = self.segment
#         s.set_style_ranges([[0,1000]], data=True)
#         for i in range(0, len(s), 25):
#             s.set_comment([[i,i+1]], "comment at %d" % i)

#         s2 = self.sub_segment
#         print(len(s2))
#         copy = s2.get_comment_locations()
#         print(copy)
#         # comments at 4 and 40 in the original means 2 and 38 in the copy
#         orig = s.get_comment_locations()
#         print(orig[0:200])
#         assert copy[2] == orig[4]
#         assert copy[28] == orig[38]

#         r = s2.get_entire_style_ranges([1], user=True)
#         print(r)
#         assert r == [((0, 23), 1), ((23, 48), 1), ((48, 73), 1), ((73, 98), 1), ((98, 123), 1), ((123, 148), 1), ((148, 173), 1), ((173, 198), 1), ((198, 200), 1)]

#     def test_split_data_at_comment2(self):
#         s = self.segment
#         start = 0
#         i = 0
#         for end in range(40, 1000, 40):
#             s.set_style_ranges([[start, end]], user=i)
#             start = end
#             i = (i + 1) % 8
#         for i in range(0, len(s), 25):
#             s.set_comment([[i,i+1]], "comment at %d" % i)

#         s2 = self.sub_segment
#         print(len(s2))
#         copy = s2.get_comment_locations()
#         print(copy)
#         # comments at 4 and 40 in the original means 2 and 38 in the copy
#         orig = s.get_comment_locations()
#         print(orig[0:200])
#         assert copy[2] == orig[4]
#         assert copy[28] == orig[38]

#         r = s2.get_entire_style_ranges([1], user=user_bit_mask)
#         print(r)
#         assert r == [((0, 38), 0), ((38, 48), 1), ((48, 73), 1), ((73, 78), 1), ((78, 118), 2), ((118, 158), 3), ((158, 198), 4), ((198, 200), 5)]

#     def test_restore_comments(self):
#         s = self.segment
#         s.set_style_ranges([[0,1000]], data=True)
#         for i in range(0, len(s), 5):
#             s.set_comment([[i,i+1]], "comment at %d" % i)

#         s1 = self.segment
#         print(len(s1))
#         indexes = [7,12]
#         r = s1.get_comment_restore_data([indexes])
#         print(r)
#         # force clear comments
#         s1.rawdata.extra.comments = {}
#         s1.style[indexes[0]:indexes[1]] = 0
#         r0 = s1.get_comment_restore_data([indexes])
#         print(r0)
#         for start, end, style, items in r0:
#             print(style)
#             assert np.all(style == 0)
#             for rawindex, comment in list(items.values()):
#                 assert not comment
#         s1.restore_comments(r)
#         r1 = s1.get_comment_restore_data([indexes])
#         print(r1)
#         for item1, item2 in zip(r, r1):
#             print(item1)
#             print(item2)
#             for a1, a2 in zip(item1, item2):
#                 print(a1, a2)
#                 if hasattr(a1, "shape"):
#                     assert np.all(a1 - a2 == 0)
#                 else:
#                     assert a1 == a2

#         s2 = self.sub_segment
#         print(len(s2))
#         indexes = [5,10]
#         r = s2.get_comment_restore_data([indexes])
#         print(r)
#         # force clear comments
#         s2.rawdata.extra.comments = {}
#         s2.style[indexes[0]:indexes[1]] = 0
#         r0 = s2.get_comment_restore_data([indexes])
#         print(r0)
#         for start, end, style, items in r0:
#             print(style)
#             assert np.all(style == 0)
#             for rawindex, comment in list(items.values()):
#                 assert not comment
#         s2.restore_comments(r)
#         r2 = s2.get_comment_restore_data([indexes])
#         print(r2)
#         for item1, item2 in zip(r, r2):
#             print(item1)
#             print(item2)
#             for a1, a2 in zip(item1, item2):
#                 print(a1, a2)
#                 if hasattr(a1, "shape"):
#                     assert np.all(a1 - a2 == 0)
#                 else:
#                     assert a1 == a2

#         for item1, item2 in zip(r1, r2):
#             print(item1)
#             print(item2)
#             # indexes won't be the same, but rawindexes and comments will
#             assert np.all(item1[2] - item2[2] == 0)
#             assert set(item1[3].values()) == set(item2[3].values())




if __name__ == "__main__":
    t = TestSegment()
    t.setup()
    t.test_metadata()
#     # t.test_indexed()
#     # t.test_indexed_sub()
#     # t.test_interleave()
#     # t = TestSegment1()
#     # t.setup()
#     # t.test_xex()
#     # t.test_copy()
#     # t = TestComments()
#     # t.setup()
#     # t.test_split_data_at_comment()
#     # t.test_restore_comments()
#     t = TestResize()
#     t.setup()
#     t.test_subset()



import numpy as np

from .. import disassembler as udis_fast

from atrip import style_bits

from .memory_map import EmptyMemoryMap

import logging
log = logging.getLogger(__name__)


# Style numbers for other disassemblers
ANTIC_DISASM = 2
JUMPMAN_LEVEL = 3
JUMPMAN_HARVEST = 4
UNINITIALIZED_DATA = 5

style_names = {
    0: "code",
    1: "data",
    ANTIC_DISASM: "antic_dl",
    JUMPMAN_LEVEL: "jumpman_level",
    JUMPMAN_HARVEST: "jumpman_harvest",
    UNINITIALIZED_DATA: "uninitialized data",
}

def get_style_name(segment, index):
    if segment.is_valid_index(index):
        s = segment.style[index] & style_bits.user_bit_mask
        msg = style_names.get(s, "")
    else:
        msg = ""
    return msg

def iter_disasm_styles():
    for i, name in style_names.items():
        if i == 0:
            continue
        yield i, name


class BaseDisassembler(object):
    name = "generic disassembler"
    cpu = "undefined"
    read_instructions = set()
    write_instructions = set()
    rw_modes = set()
    highlight_flags = 0
    default_assembler = {
        'comment char': ';',
        'origin': '*=',
        'data byte': '.byte',
        'data byte prefix': '$',
        'data byte separator': ', ',
        'name': "MAC/65",
        }

    cached_miniassemblers = {}
    label_format = "L%04X"  # Labels always upper case to match udis

    def __init__(self, asm_syntax=None, memory_map=None, hex_lower=True, mnemonic_lower=False):
        if asm_syntax is None:
            asm_syntax = self.default_assembler
        self.hex_lower = hex_lower
        self.mnemonic_lower = mnemonic_lower
        if self.hex_lower:
            self.fmt_hex2 = "%02x"
            self.fmt_hex4 = "%04x"
        else:
            self.fmt_hex2 = "%02X"
            self.fmt_hex4 = "%04X"
        if mnemonic_lower:
            case_func = lambda a:a.lower()
        else:
            case_func = lambda a:a.upper()
        self.fmt_hex_directive = asm_syntax['data byte']
        self.fmt_hex_digits = asm_syntax['data byte prefix'] + "%c%c"
        self.fmt_hex_digit_separator = asm_syntax['data byte separator']
        self.asm_origin = case_func(asm_syntax['origin'])
        self.comment_char = case_func(asm_syntax['comment char'])
        self.fast = udis_fast.DisassemblerWrapper(self.cpu, fast=True, mnemonic_lower=mnemonic_lower, hex_lower=hex_lower)
        self.memory_map = memory_map if memory_map is not None else EmptyMemoryMap()
        self.segment = None
        self.info = None
        self._pc_label_cache = None
        self._dest_pc_label_cache = None
        self._computed_directive_cache = None

    @classmethod
    def get_nop(cls):
        cpu = cputables.processors[cls.cpu]
        return cpu['nop']

    @classmethod
    def get_miniassembler(cls, cpu):
        if not cpu in cls.cached_miniassemblers:
            asm = udis_fast.MiniAssembler(cpu)
            cls.cached_miniassemblers[cpu] = asm

        return cls.cached_miniassemblers[cpu]

    @property
    def label_dict(self):
        d = {}
        if self.info:
            all_pcs = np.where(self.info.labels > 0)[0]
            inside = np.where((self.origin <= all_pcs) & (all_pcs < self.end_addr))[0]
            pcs = all_pcs[inside]
            fmt = self.label_format
            d = {pc:fmt % pc for pc in pcs}
        return d

    def assemble_text(self, pc, cmd):
        miniasm = self.get_miniassembler(self.cpu)
        byte_values = miniasm.asm(pc, cmd)
        if not byte_values:
            raise RuntimeError("Unknown addressing mode")
        return byte_values

    def add_chunk_processor(self, disassembler_name, style):
        self.fast.add_chunk_processor(disassembler_name, style)

    def disassemble_segment(self, segment):
        self.invalidate_caches()
        self.segment = segment
        self.origin = segment.origin
        self.end_addr = self.origin + len(segment)
        self.info = udis_fast.fast_disassemble_segment(self.fast, segment)
        self.use_labels = self.origin > 0
        return self.info

    def is_current(self, segment):
        return self.segment == segment and self.origin == segment.origin

    @property
    def pc_label_cache(self):
        if self._pc_label_cache is None:
            self.create_label_caches()
        return self._pc_label_cache

    @property
    def dest_pc_label_cache(self):
        if self._dest_pc_label_cache is None:
            self.create_label_caches()
        return self._dest_pc_label_cache

    @property
    def computed_directive_cache(self):
        if self._computed_directive_cache is None:
            self.create_computed_directive_cache()
        return self._computed_directive_cache

    def invalidate_caches(self):
        log.debug("Invalidating label caches")
        self._pc_label_cache = None
        self._dest_pc_label_cache = None
        self._computed_directive_cache = None

    def create_label_caches(self):
        pc_labels = {}
        dest_pc_labels = {}
        for line in self.info:
            text = self.get_pc_label(line.pc)
            if text:
                pc_labels[line.pc] = text
            if line.flag & udis_fast.flags.flag_label:
                text = self.get_dest_pc_label(line.dest_pc)
                if text:
                    dest_pc_labels[line.pc] = text
        self._pc_label_cache = pc_labels
        self._dest_pc_label_cache = dest_pc_labels
        log.debug("Created label caches: %d in pc, %d in dest_pc" % (len(pc_labels), len(dest_pc_labels)))

    def create_computed_directive_cache(self):
        pc_to_directive = {}
        for line in self.info:
            if line.flag & udis_fast.flags.flag_data_bytes:
                operand = self.get_operand_from_instruction(line.instruction)
                text = self.format_data_directive_bytes(operand)
                pc_to_directive[line.pc] = text
        self._computed_directive_cache = pc_to_directive
        log.debug("Created directive cache, %d lines" % (len(pc_to_directive)))

    def get_origin(self, pc):
        return "%s $%s" % (self.asm_origin, self.fmt_hex4 % pc)

    def get_next_instruction_pc(self, pc):
        info = self.info
        index = pc - self.origin
        row = info.index_to_row[index]
        line = info[row + 1]  # next line!
        return line.pc

    def get_comments(self, index, line=None):
        info = self.info
        if line is None:
            row = info.index_to_row[index]
            line = info[row]
        comments = []
        c = line.instruction
        if ";" in c:
            _, c = c.split(";", 1)
            comments.append(c)
        for i in range(line.num_bytes):
            c = self.segment.get_comment_at(index + i)
            if c:
                comments.append(c)
        if comments:
            return " ".join(comments)
        return ""

    def get_pc_label(self, pc):
        """Get the label for a program counter address

        """
        text = self.memory_map.get_name(pc)
        if not text and self.info.labels[pc]:
            text = self.label_format % pc
        return text

    def get_dest_pc_label(self, target_pc):
        """Get the label for a target address

        This differs from get_pc_label in that if a label doesn't exist at the
        target_pc but there is one at some small offset before, use that label
        plus the offset difference.
        """
        label = self.memory_map.get_name(target_pc)
        if not label and target_pc >= self.origin and target_pc < self.end_addr:
            #print operand, dollar, text_hex, target_pc, operand_labels_start_pc, operand_labels_end_pc
            good_opcode_target_pc = self.info.get_instruction_start_pc(target_pc)
            diff = target_pc - good_opcode_target_pc
            if diff > 0:
                # if no existing label at the target, reference it using
                # offset in bytes from the nearest previous label
                nearest_label = self.memory_map.get_name(good_opcode_target_pc)
                if not nearest_label:
                    nearest_label = self.label_format % good_opcode_target_pc
                label = "%s+%d" % (nearest_label, diff)
            else:
                label = self.label_format % target_pc
        return label

    def get_operand_label(self, line, operand):
        """Find the label that the operand points to.
        """
        target_pc = line.dest_pc

        label = self.get_dest_pc_label(target_pc)
        if label:
            # find the place to replace hex digits with the text label
            dollar = operand.find("$")
            if dollar >=0 and "#" not in operand:
                text_hex = operand[dollar+1:dollar+1+4]
                if len(text_hex) > 2 and text_hex[2] in "0123456789abcdefABCDEF":
                    size = 4
                else:
                    size = 2
                operand = operand[0:dollar] + label + operand[dollar+1+size:]
        return operand

    def get_addr_dest(self, row):
        line = self.info[row]
        return line.dest_pc if line.flag & udis_fast.flags.flag_label else -1

    def format_row_label(self, line):
        return self.fmt_hex4 % line.pc

    def format_data_list_bytes(self, index, num):
        return " ".join(self.fmt_hex2 % self.segment[index + i] for i in range(num))

    def format_data_directive_bytes(self, digits):
        """ Split string of hex digits into format used by chosen assembler

        """
        count = len(digits) // 2
        fmt = self.fmt_hex_digit_separator.join(self.fmt_hex_digits for i in range(count))
        return self.fmt_hex_directive + " " + fmt % tuple(digits[0:count*2])

    def format_label(self, line):
        if not self.use_labels:
            return "     "
        text = self.get_pc_label(line.pc)
        return "     " if not text else text

    def get_operand_from_instruction(self, text):
        if b";" in text:
            operand, _ = text.split(b";", 1)
        else:
            operand = text.rstrip()
        return operand.decode("latin-1")

    def format_operand(self, line, operand):
        if line.flag == udis_fast.flags.flag_origin:
            operand = self.get_origin(line.dest_pc)
        elif line.flag & udis_fast.flags.flag_data_bytes:
            operand = self.format_data_directive_bytes(operand)
        elif self.use_labels and line.flag & udis_fast.flags.flag_label:
            operand = self.get_operand_label(line, operand)
        return operand

    def format_instruction(self, index, line):
        label = self.format_label(line)

        # py3 notes: line.instruction is ascii byte_values, but operand will be
        # unicode after this.
        operand = self.get_operand_from_instruction(line.instruction)

        operand = self.format_operand(line, operand)
        return label + " " + operand

    def format_comment(self, index, line=None):
        info = self.info
        if line is None:
            row = info.index_to_row[index]
            line = info[row]
        comments = []
        c = line.instruction
        if b";" in c:
            _, c = c.split(b";", 1)
            comments.append(c.decode('utf-8'))
        for i in range(line.num_bytes):
            c = self.segment.get_comment_at(index + i)
            if c:
                comments.append(c)
        if comments:
            text = " ".join(comments)
            return text.replace("\r", "").replace("\n", "")
        return ""

    def iter_row_text(self, start=0, end=-1, max_bytes_per_line=8):
        """iterates over the rows representing the disassembly

        Return information designed to be used by program list formatters.
        """
        if end < 0:
            end = len(self.info.index_to_row)

        start_row = self.info.index_to_row[start]
        end_row = self.info.index_to_row[end - 1] # end is python style range, want actual last byte

        for row in range(start_row, end_row + 1):
            line = self.info[row]
            index = line.pc - self.origin
            label = self.format_label(line)
            comment = self.format_comment(index, line)
            operand = self.get_operand_from_instruction(line.instruction)
            if line.flag & udis_fast.flags.flag_data_bytes and line.num_bytes > max_bytes_per_line:
                first = True
                for i in range(0, line.num_bytes, max_bytes_per_line):
                    count = min(line.num_bytes, i + max_bytes_per_line) - i
                    hex_bytes = self.format_data_list_bytes(index + i, count)
                    subset = operand[i*2:(i+count)*2]
                    code = self.format_data_directive_bytes(subset)
                    yield line, hex_bytes, label + "   " + code, comment, count
                    if first:
                        label = "     "
                        comment = ""
                        first = False
            else:
                if line.flag == udis_fast.flags.flag_origin:
                    hex_bytes = ""
                else:
                    hex_bytes = self.format_data_list_bytes(index, line.num_bytes)
                code = self.format_operand(line, operand)
                # expand to 8 spaces
                code = label + "   " + code
                yield line, hex_bytes, code, comment, line.num_bytes

    def get_disassembled_text(self, start=0, end=-1):
        """Returns list of lines representing the disassembly

        Raises IndexError if the disassembly hasn't reached the index yet
        """
        lines = []
        start_row = self.info.index_to_row[start]
        line = self.info[start_row]
        lines.append("        " + self.get_origin(line.pc))
        for line, hex_bytes, code, comment, num_bytes in self.iter_row_text(start, end):
            if comment:
                text = "%-30s; %s" % (code, comment)
            else:
                text = code
            lines.append(text)
        return lines

    def get_atasm_lst_text(self):
        """Returns list of lines representing the disassembly

        Raises IndexError if the disassembly hasn't reached the index yet
        """
        lines = [""]
        lines.append("Source: %s.s" % (self.segment.name))
        line_num = 2
        pc = self.segment.origin
        for line, hex_bytes, code, comment, num_bytes in self.iter_row_text():
            if line.flag == udis_fast.flags.flag_origin:
                line_num += 1
                continue
            pc = line.pc
            if comment:
                code = "%-30s; %s" % (code, comment.rstrip())
            if ".byte" in code:
                count = 0
                hex_bytes = hex_bytes.upper()
                text = ""
                while count < num_bytes:
                    sub_bytes = hex_bytes[count * 3:count * 3 + 6].rstrip()
                    if count == 0:
                        text = "%05d %04X  %-17s %s" % (line_num, pc, sub_bytes, code)
                    else:
                        text += "\n%05d %04X  %s " % (line_num, pc + count, sub_bytes)
                    count += 2
            else:
                text = "%05d %04X  %-17s %s" % (line_num, line.pc, hex_bytes.upper(), code)
            lines.append(text)
            line_num += 1
        return lines

    def search_labels(self, labels, search_text, match_case=False):
        s = self.origin
        matches = []
        if not match_case:
            for pc, label in labels.items():
                if search_text in label.lower():
                    matches.append((pc - s, pc - s + 1))
        else:
            for pc, label in labels.items():
                if search_text in label:
                    matches.append((pc - s, pc - s + 1))
        return matches

    def search(self, search_text, match_case=False):
        s = self.origin
        if not match_case:
            search_text = search_text.lower()
            matches = [(t.pc - s, t.pc - s + t.num_bytes) for t in self.info if search_text in t.instruction.lower()]
        else:
            matches = [(t.pc - s, t.pc - s + t.num_bytes) for t in self.info if search_text in t.instruction]
        log.debug("instruction matches: %s" % str(matches))

        label_matches = self.search_labels(self.pc_label_cache, search_text, match_case)
        matches.extend(label_matches)
        log.debug("pc label matches: %s" % str(label_matches))

        label_matches = self.search_labels(self.dest_pc_label_cache, search_text, match_case)
        matches.extend(label_matches)
        log.debug("dest pc label matches: %s" % str(label_matches))

        label_matches = self.search_labels(self.computed_directive_cache, search_text, match_case)
        matches.extend(label_matches)
        log.debug("computed directives match: %s" % str(label_matches))

        return matches


class Basic6502Disassembler(BaseDisassembler):
    name = "6502"
    cpu = "6502"
    read_instructions = {"adc", "and", "asl", "bit", "cmp", "cpx", "cpy", "dec", "eor", "inc", "lda", "ldx", "ldy", "lsr", "ora", "rol", "ror", "sbc", "jsr", "jmp"}
    write_instructions = {"sax", "shx", "shy", "slo", "sre", "sta", "stx", "sty"}
    rw_modes = {"absolute", "absolutex", "absolutey", "indirect", "indirectx", "indirecty", "relative", "zeropage", "zeropagex", "zeropagey"}


class Undocumented6502Disassembler(Basic6502Disassembler):
    name = "6502 (with undocumented opcodes)"
    cpu = "6502undoc"


class Flagged6502Disassembler(Undocumented6502Disassembler):
    name = "6502 (highlighted undocumented opcodes)"
    # highlight_flags = udis_fast.flags.FLAG_UNDOC


class Basic65C02Disassembler(Basic6502Disassembler):
    name = "65c02"
    cpu = "65c02"


class Basic65816Disassembler(Basic6502Disassembler):
    name = "65816"
    cpu = "65816"


class Basic6800Disassembler(BaseDisassembler):
    name = "6800"
    cpu = "6800"


class Basic6809Disassembler(BaseDisassembler):
    name = "6809"
    cpu = "6809"


class Basic6811Disassembler(BaseDisassembler):
    name = "6811"
    cpu = "6811"


class Basic8051Disassembler(BaseDisassembler):
    name = "8051"
    cpu = "8051"


class Basic8080Disassembler(BaseDisassembler):
    name = "8080"
    cpu = "8080"


class BasicZ80Disassembler(BaseDisassembler):
    name = "Z80"
    cpu = "z80"



import numpy as np

from .. import errors
from ..media_type import DiskImage
from ..segment import Segment
from ..container import ContainerHeader

import logging
log = logging.getLogger(__name__)


class AtrHeader(ContainerHeader):
    # ATR Format described in http://www.atarimax.com/jindroush.atari.org/afmtatr.html
    format = np.dtype([
        ('wMagic', '<u2'),
        ('wPars', '<u2'),
        ('wSecSize', '<u2'),
        ('btParsHigh', 'u1'),
        ('dwCRC','<u4'),
        ('unused','<u4'),
        ('btFlags','u1'),
        ])
    file_format = "ATR"

    def decode_from_bytes(self, header):
        values = header.view(dtype=self.format)[0]
        if values[0] != 0x296:
            raise errors.InvalidHeader("no ATR header magic value")
        self.image_size = (int(values[3]) * 256 * 256 + int(values[1])) * 16
        self.sector_size = int(values[2])
        self.crc = int(values[4])
        self.unused = int(values[5])
        self.flags = int(values[6])

    def encode_to_bytes(self, raw):
        values = raw.view(dtype=self.format)[0]
        values[0] = 0x296
        paragraphs = self.image_size // 16
        parshigh, pars = divmod(paragraphs, 256*256)
        values[1] = pars
        values[2] = self.sector_size
        values[3] = parshigh
        values[4] = self.crc
        values[5] = self.unused
        values[6] = self.flags


class AtariSingleDensity(DiskImage):
    ui_name = "Atari SD (90K) Floppy Disk Image"
    sector_size = 128
    expected_size = 92160

    def check_header(self, header):
        if header.sector_size != self.sector_size:
            raise errors.InvalidMediaSize(f"Sector size {header.sector_size} invalid for {self.ui_name}")

    def calc_header(self, container):
        header_data = container[0:16]
        if len(header_data) == 16:
            try:
                header = AtrHeader(container)
            except errors.InvalidHeader:
                header = None
        else:
            raise errors.InvalidHeader(f"file size {len(header_data)} small to be {self.ui_name}")
        return header


class AtariSingleDensityShortImage(AtariSingleDensity):
    ui_name = "Atari SD Non-Standard Image"

    def check_disk_size(self):
        size = len(self)
        if size >= self.expected_size:
            raise errors.InvalidMediaSize(f"{self.ui_name} must be less than size {self.expected_size}")

    def check_magic(self):
        # Must have an ATR header for this to be a disk image
        if self.container.header is None:
            raise errors.InvalidHeader("Must have an ATR header for a non-standard image size")
        flag = self[0:2].view(dtype='<u2')
        if flag == 0xffff:
            raise errors.InvalidHeader("Appears to be an executable")


class AtariEnhancedDensity(AtariSingleDensity):
    ui_name = "Atari ED (130K) Floppy Disk Image"
    sector_size = 128
    expected_size = 133120


class AtariDoubleDensity(AtariSingleDensity):
    ui_name = "Atari DD (180K) Floppy Disk Image"
    sector_size = 256
    expected_size = 184320


class AtariDoubleDensityShortBootSectors(AtariDoubleDensity):
    ui_name = "Atari DD (180K) Floppy Disk Image (Short Boot Sectors)"
    expected_size = 183936
    initial_sector_size = 128
    num_initial_sectors = 3

    def calc_num_sectors(self):
        size = len(self)
        print(size)
        initial_size = self.initial_sector_size * self.num_initial_sectors
        remaining_size = size - initial_size
        if remaining_size % self.sector_size != 0:
            raise errors.InvalidMediaSize("ATR image not an integer number of sectors")
        return ((size - initial_size) // self.sector_size) + self.num_initial_sectors

    def get_index_of_sector(self, sector):
        if not self.is_sector_valid(sector):
            raise errors.ByteNotInFile166("Sector %d out of range" % sector)
        if sector <= self.num_initial_sectors:
            pos = self.num_initial_sectors * (sector - 1)
            size = self.initial_sector_size
        else:
            pos = self.num_initial_sectors * self.initial_sector_size + (sector - 1 - self.num_initial_sectors) * self.sector_size
            size = self.sector_size
        return pos, size


class AtariDoubleDensityHardDriveImage(AtariDoubleDensity):
    ui_name = "Atari DD Hard Drive Image"

    def check_disk_size(self):
        size = len(self)
        if size <= self.expected_size:
            raise errors.InvalidMediaSize(f"{self.ui_name} must be greater than size {self.expected_size}")


